import os
import logging
import ast
import sys
import re
import importlib
import importlib.machinery, importlib.util
from glob import glob
from itertools import product, chain

from contextlib import contextmanager

import yaml
import networkx as nx

from . import config

from jinja2 import Template


logger = logging.getLogger("soil")


def load_file(infile):
    folder = os.path.dirname(infile)
    if folder not in sys.path:
        sys.path.append(folder)
    with open(infile, "r") as f:
        return list(chain.from_iterable(map(expand_template, load_string(f))))


def load_string(string):
    yield from yaml.load_all(string, Loader=yaml.FullLoader)


def expand_template(config):
    if "template" not in config:
        yield config
        return
    if "vars" not in config:
        raise ValueError(
            ("You must provide a definition of variables" " for the template.")
        )

    template = config["template"]

    if not isinstance(template, str):
        template = yaml.dump(template)

    template = Template(template)

    params = params_for_template(config)

    blank_str = template.render({k: 0 for k in params[0].keys()})
    blank = list(load_string(blank_str))
    if len(blank) > 1:
        raise ValueError("Templates must not return more than one configuration")
    if "name" in blank[0]:
        raise ValueError("Templates cannot be named, use group instead")

    for ps in params:
        string = template.render(ps)
        for c in load_string(string):
            yield c


def params_for_template(config):
    sampler_config = config.get("sampler", {"N": 100})
    sampler = sampler_config.pop("method", "SALib.sample.morris.sample")
    sampler = deserializer(sampler)
    bounds = config["vars"]["bounds"]

    problem = {
        "num_vars": len(bounds),
        "names": list(bounds.keys()),
        "bounds": list(v for v in bounds.values()),
    }
    samples = sampler(problem, **sampler_config)

    lists = config["vars"].get("lists", {})
    names = list(lists.keys())
    values = list(lists.values())
    combs = list(product(*values))

    allnames = names + problem["names"]
    allvalues = [(list(i[0]) + list(i[1])) for i in product(combs, samples)]
    params = list(map(lambda x: dict(zip(allnames, x)), allvalues))
    return params


def load_files(*patterns, **kwargs):
    for pattern in patterns:
        for i in glob(pattern, **kwargs, recursive=True):
            for cfg in load_file(i):
                path = os.path.abspath(i)
                yield cfg, path


def load_config(cfg):
    if isinstance(cfg, dict):
        yield config.load_config(cfg), os.getcwd()
    else:
        yield from load_files(cfg)


_BUILTINS = None

def builtins():
    global _BUILTINS
    if not _BUILTINS:
        _BUILTINS = importlib.import_module("builtins")
    return _BUILTINS

KNOWN_MODULES = {
    'soil': None,

}

MODULE_FILES = {}

def _add_source_file(file):
    """Add a file to the list of known modules"""
    file = os.path.abspath(file)
    if file in MODULE_FILES:
        logger.warning(f"File {file} already added as module {MODULE_FILES[file]}. Reloading")
        _remove_source_file(file)
    modname = f"imported_module_{len(MODULE_FILES)}"
    loader = importlib.machinery.SourceFileLoader(modname, file)
    spec = importlib.util.spec_from_loader(loader.name, loader)
    my_module = importlib.util.module_from_spec(spec)
    loader.exec_module(my_module)
    MODULE_FILES[file] = modname
    KNOWN_MODULES[modname] = my_module

def _remove_source_file(file):
    """Remove a file from the list of known modules"""
    file = os.path.abspath(file)
    modname = None
    try:
        modname = MODULE_FILES.pop(file)
        KNOWN_MODULES.pop(modname)
    except KeyError as ex:
        raise ValueError(f"File {file} had not been added as a module: {ex}")


@contextmanager
def with_source(file=None):
    """Add a file to the list of known modules, and remove it afterwards"""
    if file:
        _add_source_file(file)
    try:
        yield
    finally:
        if file:
            _remove_source_file(file)

def get_module(modname):
    """Get a module from the list of known modules"""
    if modname not in KNOWN_MODULES or KNOWN_MODULES[modname] is None:
        module = importlib.import_module(modname)
        KNOWN_MODULES[modname] = module
    return KNOWN_MODULES[modname]


def name(value, known_modules=KNOWN_MODULES):
    """Return a name that can be imported, to serialize/deserialize an object"""
    if value is None:
        return "None"
    if not isinstance(value, type):  # Get the class name first
        value = type(value)
    tname = value.__name__
    if hasattr(builtins(), tname):
        return tname
    modname = value.__module__
    if modname == "__main__":
        return tname
    if known_modules and modname in known_modules:
        return tname
    for kmod in known_modules:
        module = get_module(kmod)
        if hasattr(module, tname):
            return tname
    return "{}.{}".format(modname, tname)


def serializer(type_):
    if type_ != "str":
        return repr
    return lambda x: x


def serialize(v, known_modules=KNOWN_MODULES):
    """Get a text representation of an object."""
    tname = name(v, known_modules=known_modules)
    func = serializer(tname)
    return func(v), tname


def serialize_dict(d, known_modules=KNOWN_MODULES):
    try:
        d = dict(d)
    except (ValueError, TypeError) as ex:
        return serialize(d)[0]
    for (k, v) in reversed(list(d.items())):
        if isinstance(v, dict):
            d[k] = serialize_dict(v, known_modules=known_modules)
        elif isinstance(v, list):
            for ix in range(len(v)):
                v[ix] = serialize_dict(v[ix], known_modules=known_modules)
        elif isinstance(v, type):
            d[k] = serialize(v, known_modules=known_modules)[1]
    return d


IS_CLASS = re.compile(r"<class '(.*)'>")


def deserializer(type_, known_modules=KNOWN_MODULES):
    if type(type_) != str:  # Already deserialized
        return type_
    if type_ == "str":
        return lambda x="": x
    if type_ == "None":
        return lambda x=None: None
    if hasattr(builtins(), type_):  # Check if it's a builtin type
        cls = getattr(builtins(), type_)
        return lambda x=None: ast.literal_eval(x) if x is not None else cls()
    match = IS_CLASS.match(type_)
    if match:
        modname, tname = match.group(1).rsplit(".", 1)
        module = get_module(modname)
        cls = getattr(module, tname)
        return getattr(cls, "deserialize", cls)

    # Otherwise, see if we can find the module and the class
    options = []

    for mod in known_modules:
        if mod:
            options.append((mod, type_))

    if "." in type_:  # Fully qualified module
        module, type_ = type_.rsplit(".", 1)
        options.append((module, type_))

    errors = []
    for modname, tname in options:
        try:
            module = get_module(modname)
            cls = getattr(module, tname)
            return getattr(cls, "deserialize", cls)
        except (ImportError, AttributeError) as ex:
            errors.append((modname, tname, ex))
    raise ValueError('Could not find type "{}". Tried: {}'.format(type_, errors))


def deserialize(type_, value=None, globs=None, **kwargs):
    """Get an object from a text representation"""
    if not isinstance(type_, str):
        return type_
    if globs and type_ in globs:
        des = globs[type_]
    else:
        try:
            des = deserializer(type_, **kwargs)
        except ValueError as ex:
            try:
                des = eval(type_)
            except Exception:
                raise ex
    if value is None:
        return des
    return des(value)


def deserialize_all(names, *args, known_modules=KNOWN_MODULES, **kwargs):
    """Return the list of deserialized objects"""
    objects = []
    for name in names:
        mod = deserialize(name, known_modules=known_modules)
        objects.append(mod(*args, **kwargs))
    return objects



from . import BaseAgent, NetworkAgent


class Ticker(BaseAgent):
    times = 0

    def step(self):
        self.times += 1

class CounterModel(NetworkAgent):
    """
    Dummy behaviour. It counts the number of nodes in the network and neighbors
    in each step and adds it to its state.
    """

    times = 0
    neighbors = 0
    total = 0

    def step(self):
        # Outside effects
        total = len(list(self.model.schedule._agents))
        neighbors = len(list(self.get_neighbors()))
        self["times"] = self.get("times", 0) + 1
        self["neighbors"] = neighbors
        self["total"] = total


class AggregatedCounter(NetworkAgent):
    """
    Dummy behaviour. It counts the number of nodes in the network and neighbors
    in each step and adds it to its state.
    """

    times = 0
    neighbors = 0
    total = 0

    def step(self):
        # Outside effects
        self["times"] += 1
        neighbors = len(list(self.get_neighbors()))
        self["neighbors"] += neighbors
        total = len(list(self.model.schedule.agents))
        self["total"] += total
        self.debug("Running for step: {}. Total: {}".format(self.now, total))



from . import Agent, state, default_state


class IndependentCascadeModel(Agent):
    """
    Settings:
        innovation_prob

        imitation_prob
    """

    time_awareness = 0
    sentimentCorrelation = 0

    # Outside effects
    @default_state
    @state
    def outside(self):
        if self.prob(self.model.innovation_prob):
            self.sentimentCorrelation = 1
            self.time_awareness = self.model.now # To know when they have been infected
            return self.imitate

    @state
    def imitate(self):
        aware_neighbors = self.get_neighbors(state_id=1, time_awareness=self.now-1)

        if self.prob(self.model.imitation_prob * len(aware_neighbors)):
            self.sentimentCorrelation = 1
            return self.outside


from __future__ import annotations

import os
import sqlite3
import math
import logging
import inspect

from typing import Any, Callable, Dict, Optional, Union, List, Type
from collections import namedtuple
from time import time as current_time
from copy import deepcopy


import networkx as nx

from mesa import Model, Agent

from . import agents as agentmod, datacollection, utils, time, network, events


# TODO: maybe add metaclass to read attributes of a model

class BaseEnvironment(Model):
    """
    The environment is key in a simulation. It controls how agents interact,
    and what information is available to them.

    This is an opinionated version of `mesa.Model` class, which adds many
    convenience methods and abstractions.

    The environment parameters and the state of every agent can be accessed
    both by using the environment as a dictionary and with the environment's
    :meth:`soil.environment.Environment.get` method.
    """

    collector_class = datacollection.SoilCollector
    schedule_class = time.TimedActivation

    def __new__(cls,
                *args: Any,
                seed="default",
                dir_path=None,
                collector_class: type = None,
                agent_reporters: Optional[Any] = None,
                model_reporters: Optional[Any] = None,
                tables: Optional[Any] = None,
                **kwargs: Any) -> Any:
        """Create a new model with a default seed value"""
        self = super().__new__(cls, *args, seed=seed, **kwargs)
        self.dir_path = dir_path or os.getcwd()
        collector_class = collector_class or cls.collector_class
        self.datacollector = collector_class(
            model_reporters=model_reporters,
            agent_reporters=agent_reporters,
            tables=tables,
        )
        for k in dir(cls):
            v = getattr(cls, k)
            if isinstance(v, property):
                v = v.fget
            if getattr(v, "add_to_report", False):
                self.add_model_reporter(k, k)

        return self

    def __init__(
        self,
        *,
        id="unnamed_env",
        seed="default",
        dir_path=None,
        schedule=None,
        schedule_class=None,
        logger = None,
        agents: Optional[Dict] = None,
        collector_class: type = datacollection.SoilCollector,
        agent_reporters: Optional[Any] = None,
        model_reporters: Optional[Any] = None,
        tables: Optional[Any] = None,
        init: bool = True,
        **env_params,
    ):

        super().__init__()


        self.current_id = -1

        self.id = id

        if logger:
            self.logger = logger
        else:
            self.logger = utils.logger.getChild(self.id)

        self.schedule = schedule
        if schedule is None:
            if schedule_class is None:
                schedule_class = self.schedule_class
            self.schedule = schedule_class(self)

        for (k, v) in env_params.items():
            self[k] = v

        if agents:
            self.add_agents(**agents)
        if init:
            self.init()
            self.datacollector.collect(self)

    def init(self):
        pass

    @property
    def get_agents(self):
        return agentmod.AgentView(self.schedule.agents)

    def agent(self, *args, **kwargs):
        return agentmod.AgentView(self.schedule.agents).one(*args, **kwargs)

    def count_agents(self, *args, **kwargs):
        return sum(1 for i in self.get_agents(*args, **kwargs))
    
    def agent_df(self, steps=False):
        df = self.datacollector.get_agent_vars_dataframe()
        if steps:
            df.index.rename(["step", "agent_id"], inplace=True)
            return df
        model_df = self.datacollector.get_model_vars_dataframe()
        df.index = df.index.set_levels(model_df.time, level=0).rename(["time", "agent_id"])
        return df

    def model_df(self, steps=False):
        df = self.datacollector.get_model_vars_dataframe()
        if steps:
            return df
        df.index.rename("step", inplace=True)
        return df.reset_index().set_index("time")

    @property
    def now(self):
        if self.schedule:
            return self.schedule.time
        raise Exception(
            "The environment has not been scheduled, so it has no sense of time"
        )

    def init_agents(self):
        pass

    def add_agent(self, agent_class, unique_id=None, **agent):
        if unique_id is None:
            unique_id = self.next_id()

        agent["unique_id"] = unique_id

        agent = dict(**agent)
        unique_id = agent.pop("unique_id", None)
        if unique_id is None:
            unique_id = self.next_id()

        a = agent_class(unique_id=unique_id, model=self, **agent)

        self.schedule.add(a)
        return a

    def add_agents(self, agent_classes: List[type], k, weights: Optional[List[float]] = None, **kwargs):
        if isinstance(agent_classes, type):
            agent_classes = [agent_classes]
        if weights is None:
            weights = [1] * len(agent_classes)

        for cls in self.random.choices(agent_classes, weights=weights, k=k):
            self.add_agent(agent_class=cls, **kwargs)

    def log(self, message, *args, level=logging.INFO, **kwargs):
        if not self.logger.isEnabledFor(level):
            return
        message = message + " ".join(str(i) for i in args)
        message = " @{:>3}: {}".format(self.now, message)
        for k, v in kwargs:
            message += " {k}={v} ".format(k, v)
        extra = {}
        extra["now"] = self.now
        extra["id"] = self.id
        return self.logger.log(level, message, extra=extra)

    def step(self):
        """
        Advance one step in the simulation, and update the data collection and scheduler appropriately
        """
        super().step()
        self.schedule.step()
        self.datacollector.collect(self)

        if self.logger.isEnabledFor(logging.DEBUG):
            msg = "Model data:\n"
            max_width = max(len(k) for k in self.datacollector.model_vars.keys())
            for (k, v) in self.datacollector.model_vars.items():
                msg += f"\t{k:<{max_width}}: {v[-1]:>6}\n"
            self.logger.debug(f"--- Steps: {self.schedule.steps:^5} - Time: {self.now:^5} --- " + msg)

    def add_model_reporter(self, name, func=None):
        if not func:
            func = name
        self.datacollector._new_model_reporter(name, func)

    def add_agent_reporter(self, name, reporter=None, agent_type=None):
        if not agent_type and not reporter:
            reporter = name
        elif agent_type:
            reporter = lambda a: reporter(a) if isinstance(a, agent_type) else None
        self.datacollector._new_agent_reporter(name, reporter)

    @classmethod
    def run(cls, *,
            name=None,
            iterations=1,
            num_processes=1, **kwargs):
        from .simulation import Simulation
        return Simulation(name=name or cls.__name__,
                          model=cls, iterations=iterations,
                          num_processes=num_processes, **kwargs).run()

    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(f"key {key}  not found in environment")

    def __delitem__(self, key):
        return delattr(self, key)

    def __contains__(self, key):
        return hasattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __str__(self):
        return str(dict(self))

    def __len__(self):
        return sum(1 for n in self.keys())

    def __iter__(self):
        return iter(self.get_agents())

    def get(self, key, default=None):
        return self[key] if key in self else default

    def keys(self):
        return (k for k in self.__dict__ if k[0] != "_")

class NetworkEnvironment(BaseEnvironment):
    """
    The NetworkEnvironment is an environment that includes one or more networkx.Graph intances
    and methods to associate agents to nodes and vice versa.
    """

    def __init__(self,
                 *args,
                 topology: Optional[Union[nx.Graph, str]] = None,
                 agent_class: Optional[Type[agentmod.Agent]] = None,
                 network_generator: Optional[Callable] = None,
                 network_params: Optional[Dict] = {},
                 init=True,
                 **kwargs):
        self.topology = topology
        self.network_generator = network_generator
        self.network_params = network_params
        if topology or network_params or network_generator:
            self.create_network(topology, generator=network_generator, **network_params)
        else:
            self.G = nx.Graph()
        super().__init__(*args, **kwargs, init=False)

        self.agent_class = agent_class
        if self.agent_class:
            self.populate_network(self.agent_class)
        self._check_agent_nodes()
        if init:
            self.init()
            self.datacollector.collect(self)

    def add_agent(self, agent_class, *args, node_id=None, topology=None, **kwargs):
        if node_id is None and topology is None:
            return super().add_agent(agent_class, *args, **kwargs)
        try:
            a = super().add_agent(agent_class, *args, node_id=node_id, **kwargs)
        except TypeError:
            self.logger.warning(f"Agent constructor for {agent_class} does not have a node_id attribute. Might be a bug.")
            a = super().add_agent(agent_class, *args, **kwargs)
        self.G.nodes[node_id]["agent"] = a
        return a

    def add_agents(self, *args, k=None, **kwargs):
        if not k and not self.G:
            raise ValueError("Cannot add agents to an empty network")
        super().add_agents(*args, k=k or len(self.G), **kwargs)

    def create_network(self, topology=None, generator=None, path=None, **network_params):
        if topology is not None:
            topology = network.from_topology(topology, dir_path=self.dir_path)
        elif path is not None:
            topology = network.from_topology(path, dir_path=self.dir_path)
        elif generator is not None:
            params = dict(generator=generator,
                                           dir_path=self.dir_path,
                                           seed=self.random,
                                           **network_params)
            try:
                topology = network.from_params(**params)
            except TypeError:
                del params["seed"]
                topology = network.from_params(**params)
        else:
            raise ValueError("topology must be a networkx.Graph or a string, or network_generator must be provided")
        self.G = topology

    def init_agents(self, *args, **kwargs):
        """Initialize the agents from a"""
        super().init_agents(*args, **kwargs)

    @property
    def network_agents(self):
        """Return agents still alive and assigned to a node in the network."""
        for (id, data) in self.G.nodes(data=True):
            if "agent" in data:
                agent = data["agent"]
                if getattr(agent, "alive", True):
                    yield agent

    def add_node(self, agent_class, unique_id=None, node_id=None, **kwargs):
        if unique_id is None:
            unique_id = self.next_id()
        if node_id is None:
            node_id = network.find_unassigned(
                G=self.G, shuffle=True, random=self.random
            )
            if node_id is None:
                node_id = f"node_for_{unique_id}"

        if node_id not in self.G.nodes:
            self.G.add_node(node_id)

        assert "agent" not in self.G.nodes[node_id]

        a = self.add_agent(
            unique_id=unique_id,
            agent_class=agent_class,
            topology=self.G,
            node_id=node_id,
            **kwargs,
        )
        a["visible"] = True
        return a

    def _check_agent_nodes(self):
        """
        Detect nodes that have agents assigned to them.
        """
        for (id, data) in self.G.nodes(data=True):
            if "agent_id" in data:
                agent = self.get_agents(data["agent_id"])
                self.G.nodes[id]["agent"] = agent
                assert not getattr(agent, "node_id", None) or agent.node_id == id
                agent.node_id = id
        for agent in self.get_agents():
            if hasattr(agent, "node_id"):
                node_id = agent["node_id"]
                if node_id not in self.G.nodes:
                    raise ValueError(f"Agent {agent} is assigned to node {agent.node_id} which is not in the network")
                node = self.G.nodes[node_id]
                if node.get("agent") is not None and node["agent"] != agent:
                    raise ValueError(f"Node {node_id} already has a different agent assigned to it")
                self.G.nodes[node_id]["agent"] = agent

    def add_agents(self, agent_classes: List[type], k=None, weights: Optional[List[float]] = None, **kwargs):
        if k is None:
            k = len(self.G)
            if not k:
                raise ValueError("Cannot add agents to an empty network")
        super().add_agents(agent_classes, k=k, weights=weights, **kwargs)

    def agent_for_node_id(self, node_id):
        return self.G.nodes[node_id].get("agent")

    def populate_network(self, agent_class: List[Model], weights: List[float] = None, **agent_params):
        if isinstance(agent_class, type):
            agent_class = [agent_class]
        else:
            agent_class = list(agent_class)
        if not weights:
            weights = [1] * len(agent_class)
        assert len(self.G)
        classes = self.random.choices(agent_class, weights, k=len(self.G))
        toadd = []
        for (cls, (node_id, node)) in zip(classes, self.G.nodes(data=True)):
            if "agent" in node:
                continue
            node["agent"] = None # Reserve
            toadd.append(dict(node_id=node_id, topology=self.G,  agent_class=cls, **agent_params))
        for d in toadd:
            a = self.add_agent(**d)
            self.G.nodes[d["node_id"]]["agent"] = a
        assert all("agent" in node for (_, node) in self.G.nodes(data=True))
        assert len(list(self.network_agents))


class EventedEnvironment(BaseEnvironment):
    def broadcast(self, msg, sender=None, expiration=None, ttl=None, **kwargs):
        for agent in self.get_agents(**kwargs):
            if agent == sender:
                continue
            self.logger.debug(f"Telling {repr(agent)}: {msg} ttl={ttl}")
            try:
                inbox = agent._inbox
            except AttributeError:
                self.logger.info(
                    f"Agent {agent.unique_id} cannot receive events because it does not have an inbox"
                )
                continue
            # Allow for AttributeError exceptions in this part of the code
            inbox.append(
                events.Tell(
                    payload=msg,
                    sender=sender,
                    expiration=expiration if ttl is None else self.now + ttl,
                )
            )


class Environment(NetworkEnvironment, EventedEnvironment):
    """Default environment class, has both network and event capabilities"""



import os
import sys
import sqlalchemy
import pandas as pd
from collections import namedtuple

def plot(env, agent_df=None, model_df=None, steps=False, ignore=["agent_count", ]):
    """Plot the model dataframe and agent dataframe together."""
    if model_df is None:
        model_df = env.model_df()
    ignore = list(ignore)
    if not steps:
        ignore.append("step")
    else:
        ignore.append("time")

    ax = model_df.drop(ignore, axis='columns').plot();
    if agent_df is None:
        try:
            agent_df = env.agent_df()
        except UserWarning:
            print("No agent dataframe provided and no agent reporters found. Skipping agent plot.", file=sys.stderr)
            return
    if not agent_df.empty:
        agent_df.unstack().apply(lambda x: x.value_counts(),
                                 axis=1).fillna(0).plot(ax=ax,
                                                        secondary_y=True)


Results = namedtuple("Results", ["config", "parameters", "env", "agents"])
#TODO implement reading from CSV

def read_sql(fpath=None, name=None, include_agents=False):
    if not (fpath is None) ^ (name is None):
        raise ValueError("Specify either a path or a simulation name")
    if name:
        fpath = os.path.join("soil_output", name, f"{name}.sqlite")
    fpath = os.path.abspath(fpath)
    # TODO: improve url parsing. This is a hacky way to check we weren't given a URL
    if "://" not in fpath:
        fpath = f"sqlite:///{fpath}"
    engine = sqlalchemy.create_engine(fpath)
    with engine.connect() as conn:
        env = pd.read_sql_table("env", con=conn,
                                index_col="step").reset_index().set_index([
                                    "simulation_id", "params_id",
                                    "iteration_id", "step"
                                ])
        agents = pd.read_sql_table("agents", con=conn, index_col=["simulation_id", "params_id", "iteration_id", "step", "agent_id"])
        config = pd.read_sql_table("configuration", con=conn, index_col="simulation_id")
        parameters = pd.read_sql_table("parameters", con=conn, index_col=["iteration_id", "params_id", "simulation_id"])
        try:
            parameters = parameters.pivot(columns="key", values="value")
        except Exception as e:
            print(f"warning: coult not pivot parameters: {e}")

        return Results(config, parameters, env, agents)



from __future__ import annotations

from typing import Dict
import os
import sys
import random

import networkx as nx

from . import config, serialization, basestring


def from_topology(topology, dir_path: str = None):
    if topology is None:
        return nx.Graph()
    if isinstance(topology, nx.Graph):
        return topology

    # If it's a dict, assume it's a node-link graph
    if isinstance(topology, dict):
        try:
            return nx.json_graph.node_link_graph(topology)
        except Exception as ex:
            raise ValueError("Unknown topology format")
    
    # Otherwise, treat like a path
    path = topology
    if dir_path and not os.path.isabs(path):
        path = os.path.join(dir_path, path)
    extension = os.path.splitext(path)[1][1:]
    kwargs = {}
    if extension == "gexf":
        kwargs["version"] = "1.2draft"
        kwargs["node_type"] = int
    try:
        method = getattr(nx.readwrite, "read_" + extension)
    except AttributeError:
        raise AttributeError("Unknown format")
    return method(path, **kwargs)


def from_params(generator, dir_path: str = None, **params):

    if dir_path not in sys.path:
        sys.path.append(dir_path)

    method = serialization.deserializer(
        generator,
        known_modules=[
            "networkx.generators",
        ],
    )
    return method(**params)


def find_unassigned(G, shuffle=False, random=random):
    """
    Link an agent to a node in a topology.

    If node_id is None, a node without an agent_id will be found.
    """
    candidates = list(G.nodes(data=True))
    if shuffle:
        random.shuffle(candidates)
    for next_id, data in candidates:
        if "agent" not in data:
            return next_id
    return None


def dump_gexf(G, f):
    for node in G.nodes():
        if "pos" in G.nodes[node]:
            G.nodes[node]["viz"] = {
                "position": {
                    "x": G.nodes[node]["pos"][0],
                    "y": G.nodes[node]["pos"][1],
                    "z": 0.0,
                }
            }
            del G.nodes[node]["pos"]

    nx.write_gexf(G, f, version="1.2draft")



import os
import logging

logger = logging.getLogger(__name__)

ROOT = os.path.dirname(__file__)
DEFAULT_FILE = os.path.join(ROOT, "VERSION")


def read_version(versionfile=DEFAULT_FILE):
    try:
        with open(versionfile) as f:
            return f.read().strip()
    except IOError:  # pragma: no cover
        logger.error(
            ("Running an unknown version of {}." "Be careful!.").format(__name__)
        )
        return "0.0"


__version__ = read_version()



from typing import Optional

import sys
import threading

import matplotlib.pyplot as plt
import reacton.ipywidgets as widgets
import solara
from solara.alias import rv

import mesa.experimental.components.matplotlib as components_matplotlib
from mesa.experimental.jupyter_viz import *
from matplotlib.figure import Figure
import networkx as nx


class Controller:
    '''
    A visualization controller that holds a reference to a model so that it can be modified or queried while the simulation is still running.
    '''
    def __init__(self):
        self.model = None


def JupyterViz(*args, **kwargs):
    c = Controller()
    page = JupyterPage(*args, controller=c, **kwargs)
    page.controller = c
    return page


@solara.component
def JupyterPage(
    model_class,
    model_params,
    controller=None,
    measures=None,
    name="Mesa Model",
    agent_portrayal=None,
    space_drawer="default",
    play_interval=150,
    columns=2,
):
    """Initialize a component to visualize a model.
    Args:
        model_class: class of the model to instantiate
        model_params: parameters for initializing the model
        measures: list of callables or data attributes to plot
        name: name for display
        agent_portrayal: options for rendering agents (dictionary)
        space_drawer: method to render the agent space for
            the model; default implementation is the `SpaceMatplotlib` component;
            simulations with no space to visualize should
            specify `space_drawer=False`
        play_interval: play interval (default: 150)
    """
    if controller is None:
        controller = Controller()

    current_step = solara.use_reactive(0)

    # 1. Set up model parameters
    user_params, fixed_params = split_model_params(model_params)
    model_parameters, set_model_parameters = solara.use_state(
        {**fixed_params, **{k: v["value"] for k, v in user_params.items()}}
    )

    # 2. Set up Model
    def make_model():
        model = model_class(**model_parameters)
        current_step.value = 0
        controller.model = model
        return model

    reset_counter = solara.use_reactive(0)
    model = solara.use_memo(
        make_model, dependencies=[*list(model_parameters.values()), reset_counter.value]
    )

    def handle_change_model_params(name: str, value: any):
        set_model_parameters({**model_parameters, name: value})

    # 3. Set up UI
    with solara.AppBar():
        solara.AppBarTitle(name)

    with solara.GridFixed(columns=2):
        UserInputs(user_params, on_change=handle_change_model_params)
        ModelController(model, play_interval, current_step, reset_counter)
        solara.Markdown(md_text=f"###Step: {current_step} - Time: {model.schedule.time } ")

    with solara.GridFixed(columns=columns):
        # 4. Space
        if space_drawer == "default":
            # draw with the default implementation
            components_matplotlib.SpaceMatplotlib(
                model, agent_portrayal, dependencies=[current_step.value]
            )
        elif space_drawer:
            # if specified, draw agent space with an alternate renderer
            space_drawer(model, agent_portrayal, dependencies=[current_step.value])
        # otherwise, do nothing (do not draw space)

        # 5. Plots
        for measure in measures:
            if callable(measure):
                # Is a custom object
                measure(model)
            else:
                components_matplotlib.PlotMatplotlib(model, measure, dependencies={current_step.value})


@solara.component
def NetworkDrawer(model, network_portrayal, dependencies: Optional[list[any]] = None):
    space_fig = Figure()
    space_ax = space_fig.subplots()
    graph = model.grid.G
    nx.draw(
        graph,
        ax=space_ax,
        **network_portrayal(graph),
    )
    solara.FigureMatplotlib(space_fig, format="png", dependencies=dependencies)


try:
    import osmnx as ox

    @solara.component
    def GeoNetworkDrawer(model, network_portrayal, dependencies: Optional[list[any]] = None):
        space_fig = Figure()
        space_ax = space_fig.subplots()
        graph = model.grid.G
        ox.plot_graph(
            graph,
            ax=space_ax,
            **network_portrayal(graph),
        )
        solara.FigureMatplotlib(space_fig, format="png", dependencies=dependencies)
except ImportError:
    pass



from __future__ import annotations

import importlib
from importlib.resources import path
import sys
import os
import logging
import traceback
from contextlib import contextmanager

from .version import __version__

try:
    basestring
except NameError:
    basestring = str

from pathlib import Path
from .agents import *
from . import agents
from .simulation import *
from .environment import Environment, EventedEnvironment
from .datacollection import SoilCollector
from . import serialization
from .utils import logger
from .time import *
from .decorators import *


def main(
    cfg="simulation.yml",
    exporters=None,
    num_processes=1,
    output="soil_output",
    *,
    debug=False,
    pdb=False,
    **kwargs,
):

    sim = None
    if isinstance(cfg, Simulation):
        sim = cfg

    import argparse
    from . import simulation

    logger.info("Running SOIL version: {}".format(__version__))

    parser = argparse.ArgumentParser(description="Run a SOIL simulation")
    parser.add_argument(
        "file",
        type=str,
        nargs="?",
        default=cfg if sim is None else "",
        help="Configuration file for the simulation (e.g., YAML or JSON)",
    )
    parser.add_argument(
        "--version", action="store_true", help="Show version info and exit"
    )
    parser.add_argument(
        "--module",
        "-m",
        type=str,
        help="file containing the code of any custom agents.",
    )
    parser.add_argument(
        "--dry-run",
        "--dry",
        action="store_true",
        help="Do not run the simulation",
    )
    parser.add_argument(
        "--no-dump",
        action="store_true",
        help="Do not store the results of the simulation to disk, show in terminal instead.",
    )
    parser.add_argument(
        "--pdb", action="store_true", help="Use a pdb console in case of exception."
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Run a customized version of a pdb console to debug a simulation.",
    )
    parser.add_argument(
        "--graph",
        "-g",
        action="store_true",
        help="Dump each iteration's network topology as a GEXF graph. Defaults to false.",
    )
    parser.add_argument(
        "--csv",
        action="store_true",
        help="Dump all data collected in CSV format. Defaults to false.",
    )
    parser.add_argument("--level", type=str, help="Logging level")
    parser.add_argument(
        "--output",
        "-o",
        type=str,
        default=output or "soil_output",
        help="folder to write results to. It defaults to the current directory.",
    )
    parser.add_argument(
        "--num-processes",
        default=num_processes,
        help="Number of processes to use for parallel execution. Defaults to 1.",
    )

    parser.add_argument(
        "-e",
        "--exporter",
        action="append",
        default=[],
        help="Export environment and/or simulations using this exporter",
    )
    parser.add_argument(
        "--max_time",
        default="-1",
        help="Set maximum time for the simulation to run. ",
    )

    parser.add_argument(
        "--max_steps",
        default="-1",
        help="Set maximum number of steps for the simulation to run.",
    )

    parser.add_argument(
        "--iterations",
        default="",
        help="Set maximum number of iterations (runs) for the simulation.",
    )

    parser.add_argument(
        "--seed",
        default=None,
        help="Manually set a seed for the simulation.",
    )

    parser.add_argument(
        "--only-convert",
        "--convert",
        action="store_true",
        help="Do not run the simulation, only convert the configuration file(s) and output them.",
    )

    parser.add_argument(
        "--set",
        metavar="KEY=VALUE",
        action="append",
        help="Set a number of parameters that will be passed to the simulation."
        "(do not put spaces before or after the = sign). "
        "If a value contains spaces, you should define "
        "it with double quotes: "
        'foo="this is a sentence". Note that '
        "values are always treated as strings.",
    )

    args = parser.parse_args()
    level = getattr(logging, (args.level or "INFO").upper())
    logger.setLevel(level)

    if args.version:
        return

    exporters = exporters or [
        "default",
    ]
    for exp in args.exporter:
        if exp not in exporters:
            exporters.append(exp)
    if args.csv:
        exporters.append("csv")
    if args.graph:
        exporters.append("gexf")

    if os.getcwd() not in sys.path:
        sys.path.append(os.getcwd())
    if args.module:
        importlib.import_module(args.module)
    if output is None:
        output = args.output

    debug = debug or args.debug

    if args.pdb or debug:
        args.synchronous = True
        os.environ["SOIL_POSTMORTEM"] = "true"

    res = []
    try:
        exp_params = {}
        opts = dict(
                    dry_run=args.dry_run,
                    dump=not args.no_dump,
                    debug=debug,
                    exporters=exporters,
                    num_processes=args.num_processes,
                    level=level,
                    outdir=output,
                    exporter_params=exp_params,
                    **kwargs)
        if args.seed is not None:
            opts["seed"] = args.seed
        if args.iterations:
            opts["iterations"] =int(args.iterations)

        if sim:
            logger.info("Loading simulation instance")
            for (k, v) in opts.items():
                setattr(sim, k, v)
            sims = [sim]
        else:
            logger.info("Loading config file: {}".format(args.file))
            if not os.path.exists(args.file):
                logger.error("Please, input a valid file")
                return

            assert opts["debug"] == debug
            sims = list(
                simulation.iter_from_file(
                    args.file,
                    **opts,
                )
            )

        for sim in sims:
            assert sim.debug == debug

            if args.set:
                for s in args.set:
                    k, v = s.split("=", 1)[:2]
                    v = eval(v)
                    tail, *head = k.rsplit(".", 1)[::-1]
                    target = sim.parameters
                    if head:
                        for part in head[0].split("."):
                            try:
                                target = getattr(target, part)
                            except AttributeError:
                                target = target[part]
                    try:
                        setattr(target, tail, v)
                    except AttributeError:
                        target[tail] = v

            if args.only_convert:
                print(sim.to_yaml())
                continue
            max_time = float(args.max_time) if args.max_time != "-1" else None
            max_steps = float(args.max_steps) if args.max_steps != "-1" else None
            res.append(sim.run(max_time=max_time, max_steps=max_steps))

    except Exception as ex:
        if args.pdb:
            from .debugging import post_mortem

            print(traceback.format_exc())
            post_mortem()
        else:
            raise
    if debug:
        from .debugging import set_trace

        os.environ["SOIL_DEBUG"] = "true"
        set_trace()
    return res


@contextmanager
def easy(cfg, pdb=False, debug=False, **kwargs):
    try:
        return main(cfg, debug=debug, pdb=pdb, **kwargs)[0]
    except Exception as e:
        if os.environ.get("SOIL_POSTMORTEM"):
            from .debugging import post_mortem

            print(traceback.format_exc())
            post_mortem()
        raise


if __name__ == "__main__":
    main()



from scipy.spatial import cKDTree as KDTree
import networkx as nx
from . import NetworkAgent


class Geo(NetworkAgent):
    """In this type of network, nodes have a "pos" attribute."""

    def geo_search(self, radius, center=False, **kwargs):
        """Get a list of nodes whose coordinates are closer than *radius* to *node*."""
        node = self.node_id

        G = self.subgraph(**kwargs)

        pos = nx.get_node_attributes(G, "pos")
        if not pos:
            return []
        nodes, coords = list(zip(*pos.items()))
        kdtree = KDTree(coords)  # Cannot provide generator.
        indices = kdtree.query_ball_point(pos[node], radius)
        return [nodes[i] for i in indices if center or (nodes[i] != node)]


import io
import threading
import asyncio
import logging
import networkx as nx
import os
import sys
import tornado.ioloop
import tornado.web
import tornado.websocket
import tornado.escape
import tornado.gen
import yaml
import webbrowser
from contextlib import contextmanager
from time import sleep
from xml.etree.ElementTree import tostring

from tornado.concurrent import run_on_executor
from concurrent.futures import ThreadPoolExecutor

from ..simulation import Simulation

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

ROOT = os.path.abspath(os.path.dirname(__file__))

MAX_WORKERS = 4
LOGGING_INTERVAL = 0.5

# Workaround to let Soil load the required modules
sys.path.append(ROOT)


class PageHandler(tornado.web.RequestHandler):
    """Handler for the HTML template which holds the visualization."""

    def get(self):
        self.render(
            "index.html", port=self.application.port, name=self.application.name
        )


class SocketHandler(tornado.websocket.WebSocketHandler):
    """Handler for websocket."""

    executor = ThreadPoolExecutor(max_workers=MAX_WORKERS)

    def open(self):
        if self.application.verbose:
            logger.info("Socket opened!")

    def check_origin(self, origin):
        return True

    def on_message(self, message):
        """Receiving a message from the websocket, parse, and act accordingly."""

        msg = tornado.escape.json_decode(message)

        if msg["type"] == "config_file":

            if self.application.verbose:
                print(msg["data"])

            self.config = list(yaml.load_all(msg["data"]))

            if len(self.config) > 1:
                error = "Please, provide only one configuration."
                if self.application.verbose:
                    logger.error(error)
                self.write_message({"type": "error", "error": error})
                return

            self.config = self.config[0]
            self.send_log(
                "INFO." + self.simulation_name,
                "Using config: {name}".format(name=self.config["name"]),
            )

            if "visualization_params" in self.config:
                self.write_message(
                    {
                        "type": "visualization_params",
                        "data": self.config["visualization_params"],
                    }
                )
            self.name = self.config["name"]
            self.run_simulation()

            settings = []
            for key in self.config["environment_params"]:
                if (
                    type(self.config["environment_params"][key]) == float
                    or type(self.config["environment_params"][key]) == int
                ):
                    if self.config["environment_params"][key] <= 1:
                        setting_type = "number"
                    else:
                        setting_type = "great_number"
                elif type(self.config["environment_params"][key]) == bool:
                    setting_type = "boolean"
                else:
                    setting_type = "undefined"

                settings.append(
                    {
                        "label": key,
                        "type": setting_type,
                        "value": self.config["environment_params"][key],
                    }
                )

            self.write_message({"type": "settings", "data": settings})

        elif msg["type"] == "get_trial":
            if self.application.verbose:
                logger.info("Trial {} requested!".format(msg["data"]))
            self.send_log("INFO." + __name__, "Trial {} requested!".format(msg["data"]))
            self.write_message(
                {"type": "get_trial", "data": self.get_trial(int(msg["data"]))}
            )

        elif msg["type"] == "run_simulation":
            if self.application.verbose:
                logger.info(
                    "Running new simulation for {name}".format(name=self.config["name"])
                )
            self.send_log(
                "INFO." + self.simulation_name,
                "Running new simulation for {name}".format(name=self.config["name"]),
            )
            self.config["environment_params"] = msg["data"]
            self.run_simulation()

        elif msg["type"] == "download_gexf":
            G = self.trials[int(msg["data"])].history_to_graph()
            for node in G.nodes():
                if "pos" in G.nodes[node]:
                    G.nodes[node]["viz"] = {
                        "position": {
                            "x": G.nodes[node]["pos"][0],
                            "y": G.nodes[node]["pos"][1],
                            "z": 0.0,
                        }
                    }
                    del G.nodes[node]["pos"]
            writer = nx.readwrite.gexf.GEXFWriter(version="1.2draft")
            writer.add_graph(G)
            self.write_message(
                {
                    "type": "download_gexf",
                    "filename": self.config["name"] + "_trial_" + str(msg["data"]),
                    "data": tostring(writer.xml).decode(writer.encoding),
                }
            )

        elif msg["type"] == "download_json":
            G = self.trials[int(msg["data"])].history_to_graph()
            for node in G.nodes():
                if "pos" in G.nodes[node]:
                    G.nodes[node]["viz"] = {
                        "position": {
                            "x": G.nodes[node]["pos"][0],
                            "y": G.nodes[node]["pos"][1],
                            "z": 0.0,
                        }
                    }
                    del G.nodes[node]["pos"]
            self.write_message(
                {
                    "type": "download_json",
                    "filename": self.config["name"] + "_trial_" + str(msg["data"]),
                    "data": nx.node_link_data(G),
                }
            )

        else:
            if self.application.verbose:
                logger.info("Unexpected message!")

    def update_logging(self):
        try:
            if (
                not self.log_capture_string.closed
                and self.log_capture_string.getvalue()
            ):
                for i in range(len(self.log_capture_string.getvalue().split("\n")) - 1):
                    self.send_log(
                        "INFO." + self.simulation_name,
                        self.log_capture_string.getvalue().split("\n")[i],
                    )
                self.log_capture_string.truncate(0)
                self.log_capture_string.seek(0)
        finally:
            if self.capture_logging:
                tornado.ioloop.IOLoop.current().call_later(
                    LOGGING_INTERVAL, self.update_logging
                )

    def on_close(self):
        if self.application.verbose:
            logger.info("Socket closed!")

    def send_log(self, logger, logging):
        self.write_message({"type": "log", "logger": logger, "logging": logging})

    @property
    def simulation_name(self):
        return self.config.get("name", "NoSimulationRunning")

    @run_on_executor
    def nonblocking(self, config):
        simulation = Simulation(**config)
        return simulation.run()

    @tornado.gen.coroutine
    def run_simulation(self):
        # Run simulation and capture logs
        logger.info("Running simulation!")
        if "visualization_params" in self.config:
            del self.config["visualization_params"]
        with self.logging(self.simulation_name):
            try:
                config = dict(**self.config)
                config["outdir"] = os.path.join(self.application.outdir, config["name"])
                config["dump"] = self.application.dump
                self.trials = yield self.nonblocking(config)

                self.write_message(
                    {
                        "type": "trials",
                        "data": list(trial.name for trial in self.trials),
                    }
                )
            except Exception as ex:
                error = "Something went wrong:\n\t{}".format(ex)
                logging.info(error)
                self.write_message({"type": "error", "error": error})
                self.send_log("ERROR." + self.simulation_name, error)

    def get_trial(self, trial):
        logger.info("Available trials: %s " % len(self.trials))
        logger.info("Ask for : %s" % trial)
        trial = self.trials[trial]
        G = trial.history_to_graph()
        return nx.node_link_data(G)

    @contextmanager
    def logging(self, logger):
        self.capture_logging = True
        self.logger_application = logging.getLogger(logger)
        self.log_capture_string = io.StringIO()
        ch = logging.StreamHandler(self.log_capture_string)
        self.logger_application.addHandler(ch)
        self.update_logging()
        yield self.capture_logging

        sleep(0.2)
        self.log_capture_string.close()
        self.logger_application.removeHandler(ch)
        self.capture_logging = False
        return self.capture_logging


class ModularServer(tornado.web.Application):
    """Main visualization application."""

    port = 8001
    page_handler = (r"/", PageHandler)
    socket_handler = (r"/ws", SocketHandler)
    static_handler = (
        r"/(.*)",
        tornado.web.StaticFileHandler,
        {"path": os.path.join(ROOT, "static")},
    )
    local_handler = (r"/local/(.*)", tornado.web.StaticFileHandler, {"path": ""})

    handlers = [page_handler, socket_handler, static_handler, local_handler]
    settings = {"debug": True, "template_path": ROOT + "/templates"}

    def __init__(
        self, dump=False, outdir="output", name="SOIL", verbose=True, *args, **kwargs
    ):

        self.verbose = verbose
        self.name = name
        self.dump = dump
        self.outdir = outdir

        # Initializing the application itself:
        super().__init__(self.handlers, **self.settings)

    def launch(self, port=None):
        """Run the app."""

        if port is not None:
            self.port = port
        url = "http://127.0.0.1:{PORT}".format(PORT=self.port)
        print("Interface starting at {url}".format(url=url))
        self.listen(self.port)
        # webbrowser.open(url)
        tornado.ioloop.IOLoop.instance().start()


def run(*args, **kwargs):
    asyncio.set_event_loop(asyncio.new_event_loop())
    server = ModularServer(*args, **kwargs)
    server.launch()


def main():
    import argparse

    parser = argparse.ArgumentParser(description="Visualization of a Graph Model")

    parser.add_argument(
        "--name", "-n", nargs=1, default="SOIL", help="name of the simulation"
    )
    parser.add_argument(
        "--dump", "-d", help="dumping results in folder output", action="store_true"
    )
    parser.add_argument(
        "--port", "-p", nargs=1, default=8001, help="port for launching the server"
    )
    parser.add_argument("--verbose", "-v", help="verbose mode", action="store_true")
    args = parser.parse_args()

    run(
        name=args.name,
        port=(args.port[0] if isinstance(args.port, list) else args.port),
        verbose=args.verbose,
    )



from . import FSM, state, default_state


class BassModel(FSM):
    """
    Settings:
        innovation_prob
        imitation_prob
    """

    sentimentCorrelation = 0

    def step(self):
        self.behaviour()

    @default_state
    @state
    def innovation(self):
        if self.prob(self.innovation_prob):
            self.sentimentCorrelation = 1
            return self.aware
        else:
            aware_neighbors = self.get_neighbors(state_id=self.aware.id)
            num_neighbors_aware = len(aware_neighbors)
            if self.prob((self.imitation_prob * num_neighbors_aware)):
                self.sentimentCorrelation = 1
                return self.aware

    @state
    def aware(self):
        self.die()


from mesa import DataCollector as MDC


class SoilCollector(MDC):
    def __init__(self, model_reporters=None, agent_reporters=None, tables=None, **kwargs):
        model_reporters = model_reporters or {}
        agent_reporters = agent_reporters or {}
        tables = tables or {}
        if 'agent_count' not in model_reporters:
            model_reporters['agent_count'] = lambda m: m.schedule.get_agent_count()
        if 'time' not in model_reporters:
            model_reporters['time'] = lambda m: m.schedule.time
        # if 'state_id' not in agent_reporters:
        #     agent_reporters['state_id'] = lambda agent: getattr(agent, 'state_id', None)

        super().__init__(model_reporters=model_reporters,
                         agent_reporters=agent_reporters,
                         tables=tables,
                         **kwargs)



import os
import sys
from time import time as current_time
from io import BytesIO
from textwrap import dedent, indent


import networkx as nx
import pandas as pd


from .serialization import deserialize, serialize
from .utils import try_backup, open_or_reuse, logger, timer


from . import utils, network


class DryRunner(BytesIO):
    def __init__(self, fname, *args, copy_to=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.__fname = fname
        self.__copy_to = copy_to

    def write(self, txt):
        if self.__copy_to:
            self.__copy_to.write("{}:::{}".format(self.__fname, txt))
        try:
            super().write(txt)
        except TypeError:
            super().write(bytes(txt, "utf-8"))

    def close(self):
        content = "(binary data not shown)"
        try:
            content = self.getvalue().decode()
        except UnicodeDecodeError:
            pass
        logger.info(
            "**Not** written to {} (no_dump mode):\n\n{}\n\n".format(
                self.__fname, content
            )
        )
        super().close()


class Exporter:
    """
    Interface for all exporters. It is not necessary, but it is useful
    if you don't plan to implement all the methods.
    """

    def __init__(self, simulation, outdir=None, dump=True, copy_to=None):
        self.simulation = simulation
        outdir = outdir or os.path.join(os.getcwd(), "soil_output")
        self.outdir = os.path.join(outdir, simulation.group or "", simulation.name)
        self.dump = dump
        if copy_to is None and not dump:
            copy_to = sys.stdout
        self.copy_to = copy_to

    def sim_start(self):
        """Method to call when the simulation starts"""
        pass

    def sim_end(self):
        """Method to call when the simulation ends"""
        pass

    def iteration_start(self, env):
        """Method to call when a iteration start"""
        pass

    def iteration_end(self, env, params, params_id):
        """Method to call when a iteration ends"""
        pass

    def output(self, f, mode="w", **kwargs):
        if not self.dump:
            f = DryRunner(f, copy_to=self.copy_to)
        else:
            try:
                if not os.path.isabs(f):
                    f = os.path.join(self.outdir, f)
            except TypeError:
                pass
        return open_or_reuse(f, mode=mode, backup=self.simulation.backup, **kwargs)

    def get_dfs(self, env, **kwargs):
        yield from get_dc_dfs(env.datacollector,
                              simulation_id=self.simulation.id,
                              iteration_id=env.id,
                              **kwargs)


def get_dc_dfs(dc, **kwargs):
    dfs = {}
    dfe = dc.get_model_vars_dataframe()
    dfe.index.rename("step", inplace=True)
    dfs["env"] = dfe
    try:
        dfa = dc.get_agent_vars_dataframe() 
        dfa.index.rename(["step", "agent_id"], inplace=True)
        dfs["agents"] = dfa
    except UserWarning:
        pass
    for table_name in dc.tables:
        dfs[table_name] = dc.get_table_dataframe(table_name)
    for (name, df) in dfs.items():
        for (k, v) in kwargs.items():
            df[k] = v
        df.set_index(["simulation_id", "iteration_id"], append=True, inplace=True)
    
    yield from dfs.items()


class SQLite(Exporter):
    """Writes sqlite results"""
    sim_started = False

    def sim_start(self):
        if not self.dump:
            logger.debug("NOT dumping results")
            return

        from sqlalchemy import create_engine

        self.dbpath = os.path.join(self.outdir, f"{self.simulation.name}.sqlite")
        logger.info("Dumping results to %s", self.dbpath)
        if self.simulation.backup:
            try_backup(self.dbpath, remove=True)
        
        if self.simulation.overwrite:
            if os.path.exists(self.dbpath):
                os.remove(self.dbpath)
        
        self.engine = create_engine(f"sqlite:///{self.dbpath}", echo=False)

        sim_dict = {k: serialize(v)[0] for (k,v) in self.simulation.to_dict().items()}
        sim_dict["simulation_id"] = self.simulation.id
        df = pd.DataFrame([sim_dict])
        df.to_sql("configuration", con=self.engine, if_exists="append")

    def iteration_end(self, env, params, params_id, *args, **kwargs):
        if not self.dump:
            logger.info("Running in NO DUMP mode. Results will NOT be saved to a DB.")
            return

        with timer(
            "Dumping simulation {} iteration {}".format(self.simulation.name, env.id)
        ):

            pd.DataFrame([{"simulation_id": self.simulation.id,
                           "params_id": params_id,
                           "iteration_id": env.id,
                           "key": k,
                           "value": serialize(v)[0]} for (k,v) in params.items()]).to_sql("parameters", con=self.engine, if_exists="append")

            for (t, df) in self.get_dfs(env, params_id=params_id):
                df.to_sql(t, con=self.engine, if_exists="append")

class csv(Exporter):
    """Export the state of each environment (and its agents) a CSV file for the simulation"""

    def sim_start(self):
        super().sim_start()

    def iteration_end(self, env, params, params_id, *args, **kwargs):
        with timer(
            "[CSV] Dumping simulation {} iteration {} @ dir {}".format(
                self.simulation.name, env.id, self.outdir
            )
        ):
            for (df_name, df) in self.get_dfs(env, params_id=params_id):
                with self.output("{}.{}.csv".format(env.id, df_name), mode="a") as f:
                    df.to_csv(f)


class gexf(Exporter):
    def iteration_end(self, env, *args, **kwargs):
        if not self.dump:
            logger.info("Not dumping GEXF (NO_DUMP mode)")
            return

        with timer(
            "[GEXF] Dumping simulation {} iteration {}".format(self.simulation.name, env.id)
        ):
            with self.output("{}.gexf".format(env.id), mode="wb") as f:
                nx.write_gexf(env.G, f)


class dummy(Exporter):
    def sim_start(self):
        with self.output("dummy", "w") as f:
            f.write("simulation started @ {}\n".format(current_time()))

    def iteration_start(self, env):
        with self.output("dummy", "w") as f:
            f.write("iteration started@ {}\n".format(current_time()))

    def iteration_end(self, env, *args, **kwargs):
        with self.output("dummy", "w") as f:
            f.write("iteration ended@ {}\n".format(current_time()))

    def sim_end(self):
        with self.output("dummy", "a") as f:
            f.write("simulation ended @ {}\n".format(current_time()))


class graphdrawing(Exporter):
    def iteration_end(self, env, *args, **kwargs):
        import matplotlib.pyplot as plt
        # Outside effects
        f = plt.figure()
        nx.draw(
            env.G,
            node_size=10,
            width=0.2,
            pos=nx.spring_layout(env.G, scale=100),
            ax=f.add_subplot(111),
        )
        with open("graph-{}.png".format(env.id)) as f:
            f.savefig(f)


class summary(Exporter):
    """Print a summary of each iteration to sys.stdout"""

    def iteration_end(self, env, *args, **kwargs):
        msg = ""
        for (t, df) in self.get_dfs(env):
            if not len(df):
                continue
            tabs = "\t" * 2
            description = indent(str(df.describe()), tabs)
            last_line = indent(str(df.iloc[-1:]), tabs)
            # value_counts = indent(str(df.value_counts()), tabs)
            value_counts = indent(str(df.apply(lambda x: x.value_counts()).T.stack()), tabs)

            msg += dedent("""
            Dataframe {t}:
                Last line: :
            {last_line}

                Description:
            {description}

                Value counts:
            {value_counts}

            """).format(**locals())
        logger.info(msg)

class YAML(Exporter):
    """Writes the configuration of the simulation to a YAML file"""

    def sim_start(self):
        if not self.dump:
            logger.debug("NOT dumping results")
            return
        with self.output(self.simulation.id + ".dumped.yml") as f:
            logger.info(f"Dumping simulation configuration to {self.outdir}")
            f.write(self.simulation.to_yaml())

class default(Exporter):
    """Default exporter. Writes sqlite results, as well as the simulation YAML"""

    def __init__(self, *args, exporter_cls=[], **kwargs):
        exporter_cls = exporter_cls or [YAML, SQLite]
        self.inner = [cls(*args, **kwargs) for cls in exporter_cls]

    def sim_start(self, *args, **kwargs):
        for exporter in self.inner:
            exporter.sim_start(*args, **kwargs)

    def sim_end(self, *args, **kwargs):
        for exporter in self.inner:
            exporter.sim_end(*args, **kwargs)

    def iteration_end(self, *args, **kwargs):
        for exporter in self.inner:
            exporter.iteration_end(*args, **kwargs)



from . import BaseAgent
from ..events import Message, Tell, Ask, TimedOut
from functools import partial
from collections import deque
from types import coroutine


class EventedAgent(BaseAgent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._inbox = deque()
        self._processed = 0

    def on_receive(self, *args, **kwargs):
        pass

    @coroutine
    def received(self, expiration=None, timeout=60, delay=1, process=True):
        if not expiration:
            expiration = self.now + timeout
        while self.now < expiration:
            if self._inbox:
                msgs = self._inbox
                if process:
                    self.process_messages()
                return msgs
            yield self.delay(delay)
        raise TimedOut("No message received")

    def tell(self, msg, sender=None):
        self._inbox.append(Tell(timestamp=self.now, payload=msg, sender=sender))

    @coroutine
    def ask(self, msg, expiration=None, timeout=None, delay=1):
        ask = Ask(timestamp=self.now, payload=msg, sender=self)
        self._inbox.append(ask)
        expiration = float("inf") if timeout is None else self.now + timeout
        while self.now < expiration:
            if ask.reply:
                return ask.reply
            yield self.delay(delay)
        raise TimedOut("No reply received")

    def process_messages(self):
        valid = list()
        for msg in self._inbox:
            self._processed += 1
            if msg.expired(self.now):
                continue
            valid.append(msg)
            reply = self.on_receive(msg.payload, sender=msg.sender)
            if isinstance(msg, Ask):
                msg.reply = reply
        self._inbox.clear()
        return valid


Evented = EventedAgent



import logging
from time import time as current_time, strftime, gmtime, localtime
import os
import traceback

from functools import partial
from shutil import copyfile, move
from multiprocessing import Pool, cpu_count

from contextlib import contextmanager

logger = logging.getLogger("soil")
logger.setLevel(logging.WARNING)

timeformat = "%H:%M:%S"

if os.environ.get("SOIL_VERBOSE", ""):
    logformat = "[%(levelname)-5.5s][%(asctime)s][%(name)s]:  %(message)s"
else:
    logformat = "[%(levelname)-5.5s][%(asctime)s] %(message)s"

logFormatter = logging.Formatter(logformat, timeformat)
consoleHandler = logging.StreamHandler()
consoleHandler.setFormatter(logFormatter)

logging.basicConfig(
    level=logging.WARNING,
    handlers=[
        consoleHandler,
    ],
)


@contextmanager
def timer(name="task", pre="", function=logger.info, to_object=None):
    start = current_time()
    function("{}Starting {} at {}.".format(pre, name, strftime("%X", gmtime(start))))
    yield start
    end = current_time()
    function(
        "{}Finished {} at {} in {} seconds".format(
            pre, name, strftime("%X", gmtime(end)), str(end - start)
        )
    )
    if to_object:
        to_object.start = start
        to_object.end = end


def try_backup(path, remove=False):
    if not os.path.exists(path):
        return None
    outdir = os.path.dirname(path)
    if outdir and not os.path.exists(outdir):
        os.makedirs(outdir)
    creation = os.path.getctime(path)
    stamp = strftime("%Y-%m-%d_%H.%M.%S", localtime(creation))

    backup_dir = os.path.join(outdir, "backup")
    if not os.path.exists(backup_dir):
        os.makedirs(backup_dir)
    newpath = os.path.join(backup_dir, "{}@{}".format(os.path.basename(path), stamp))
    if remove:
        move(path, newpath)
    else:
        copyfile(path, newpath)
    return newpath


def safe_open(path, mode="r", backup=True, **kwargs):
    outdir = os.path.dirname(path)
    if outdir and not os.path.exists(outdir):
        os.makedirs(outdir)
    if backup and "w" in mode:
        try_backup(path)
    return open(path, mode=mode, **kwargs)


@contextmanager
def open_or_reuse(f, *args, **kwargs):
    try:
        with safe_open(f, *args, **kwargs) as f:
            yield f
    except (AttributeError, TypeError) as ex:
        yield f


def flatten_dict(d):
    if not isinstance(d, dict):
        return d
    return dict(_flatten_dict(d))


def _flatten_dict(d, prefix=""):
    if not isinstance(d, dict):
        # print('END:', prefix, d)
        yield prefix, d
        return
    if prefix:
        prefix = prefix + "."
    for k, v in d.items():
        # print(k, v)
        res = list(_flatten_dict(v, prefix="{}{}".format(prefix, k)))
        # print('RES:', res)
        yield from res


def unflatten_dict(d):
    out = {}
    for k, v in d.items():
        target = out
        if not isinstance(k, str):
            target[k] = v
            continue
        tokens = k.split(".")
        if len(tokens) < 2:
            target[k] = v
            continue
        for token in tokens[:-1]:
            if token not in target:
                target[token] = {}
            target = target[token]
        target[tokens[-1]] = v
    return out


def run_and_return_exceptions(func, *args, **kwargs):
    """
    A wrapper for a function that catches exceptions and returns them.
    It is meant for async simulations.
    """
    try:
        return func(*args, **kwargs)
    except Exception as ex:
        if ex.__cause__ is not None:
            ex = ex.__cause__
        ex.message = "".join(
            traceback.format_exception(type(ex), ex, ex.__traceback__)[:]
        )
        return ex


def run_parallel(func, iterable, num_processes=1, **kwargs):
    if num_processes > 1 and not os.environ.get("SOIL_DEBUG", None):
        if num_processes < 1:
            num_processes = cpu_count() - num_processes
        p = Pool(processes=num_processes)
        wrapped_func = partial(run_and_return_exceptions, func, **kwargs)
        for i in p.imap_unordered(wrapped_func, iterable):
            if isinstance(i, Exception):
                logger.error("Trial failed:\n\t%s", i.message)
                continue
            yield i
    else:
        for i in iterable:
            yield func(i, **kwargs)


def int_seed(seed: str):
    return int.from_bytes(seed.encode(), "little")


import os
from time import time as current_time, strftime
import sys
import yaml
import hashlib

import inspect
import logging
import networkx as nx

from tqdm.auto import tqdm

from textwrap import dedent

from dataclasses import dataclass, field, asdict, replace
from typing import Any, Dict, Union, Optional, List


from functools import partial
from contextlib import contextmanager
from itertools import product
import json


from . import serialization, exporters, utils, basestring, agents
from . import environment
from .utils import logger, run_and_return_exceptions
from .debugging import set_trace

_AVOID_RUNNING = False
_QUEUED = []

@contextmanager
def do_not_run(): 
    global _AVOID_RUNNING
    _AVOID_RUNNING = True
    try:
        logger.debug("NOT RUNNING")
        yield
    finally:
        logger.debug("RUNNING AGAIN")
        _AVOID_RUNNING = False


def _iter_queued():
    while _QUEUED:
        (cls, params) = _QUEUED.pop(0)
        yield replace(cls, parameters=params)


# TODO: change documentation for simulation
@dataclass
class Simulation:
    """
    A simulation is a collection of agents and a model. It is responsible for running the model and agents, and collecting data from them.

    Args:
        version: The version of the simulation. This is used to determine how to load the simulation.
        name: The name of the simulation.
        description: A description of the simulation.
        group: The group that the simulation belongs to.
        model: The model to use for the simulation. This can be a string or a class.
        parameters: The parameters to pass to the model.
        matrix: A matrix of values for each parameter.
        seed: The seed to use for the simulation.
        dir_path: The directory path to use for the simulation.
        max_time: The maximum time to run the simulation.
        max_steps: The maximum number of steps to run the simulation.
        iterations: The number of iterations (times) to run the simulation.
        num_processes: The number of processes to use for the simulation. If greater than one, simulations will be performed in parallel. This may make debugging and error handling difficult.
        tables: The tables to use in the simulation datacollector
        agent_reporters: The agent reporters to use in the datacollector
        model_reporters: The model reporters to use in the datacollector
        dry_run: Whether or not to run the simulation. If True, the simulation will not be run.
        backup: Whether or not to backup the simulation. If True, the simulation files will be backed up to a different directory.
        overwrite: Whether or not to replace existing simulation data.
        source_file: Python file to use to find additional classes.
    """

    version: str = "2"
    source_file: Optional[str] = None
    name: Optional[str] = None
    description: Optional[str] = ""
    group: str = None
    backup: bool = False
    overwrite: bool = False
    dry_run: bool = False
    dump: bool = False
    model: Union[str, type] = "soil.Environment"
    parameters: dict = field(default_factory=dict)
    matrix: dict = field(default_factory=dict)
    seed: str = "default"
    dir_path: str = field(default_factory=lambda: os.getcwd())
    max_time: float = None
    max_steps: int = None
    iterations: int = 1
    num_processes: Optional[int] = 1
    exporters: Optional[List[str]] = field(default_factory=lambda: [exporters.default])
    model_reporters: Optional[Dict[str, Any]] = field(default_factory=dict)
    agent_reporters: Optional[Dict[str, Any]] = field(default_factory=dict)
    tables: Optional[Dict[str, Any]] = field(default_factory=dict)
    outdir: str = field(default_factory=lambda: os.path.join(os.getcwd(), "soil_output"))
    # outdir: Optional[str] = None
    exporter_params: Optional[Dict[str, Any]] = field(default_factory=dict)
    level: int = logging.INFO
    skip_test: Optional[bool] = False
    debug: Optional[bool] = False

    def __post_init__(self):
        if self.name is None:
            if isinstance(self.model, str):
                self.name = self.model
            else:
                self.name = self.model.__name__
        self.logger = logger.getChild(self.name)
        self.logger.setLevel(self.level)

        if self.source_file and (not os.path.isabs(self.source_file)):
            self.source_file = os.path.abspath(os.path.join(self.dir_path, self.source_file))
        with serialization.with_source(self.source_file):

            if isinstance(self.model, str):
                self.model = serialization.deserialize(self.model)

            self.agent_reporters = self.agent_reporters
            self.model_reporters = self.model_reporters
            self.tables = self.tables
            self.id = f"{self.name}_{current_time()}"

    def run(self, **kwargs):
        """Run the simulation and return the list of resulting environments"""
        if kwargs:
            return replace(self, **kwargs).run()

        param_combinations = self._collect_params(**kwargs)
        if _AVOID_RUNNING:
            _QUEUED.extend((self, param) for param in param_combinations)
            return []

        self.logger.debug("Using exporters: %s", self.exporters or [])

        exporters = serialization.deserialize_all(
            self.exporters,
            simulation=self,
            known_modules=[
                "soil.exporters",
            ],
            dump=self.dump and not self.dry_run,
            outdir=self.outdir,
            **self.exporter_params,
        )

        results = []
        for exporter in exporters:
            exporter.sim_start()

        for params in tqdm(param_combinations, desc=self.name, unit="configuration"):
            tqdm.write("- Running for parameters: ")
            for (k, v) in params.items():
                tqdm.write(f"  {k} = {v}")
            sha = hashlib.sha256()
            sha.update(repr(sorted(params.items())).encode())
            params_id = sha.hexdigest()[:7]
            for env in self._run_iters_for_params(params):
                for exporter in exporters:
                    exporter.iteration_end(env, params, params_id)
                results.append(env)

        for exporter in exporters:
            exporter.sim_end()

        return results

    def _collect_params(self):

        parameters = []
        if self.parameters:
            parameters.append(self.parameters)
        if self.matrix:
            assert isinstance(self.matrix, dict)
            for values in product(*(self.matrix.values())):
                parameters.append(dict(zip(self.matrix.keys(), values)))

        if not parameters:
            parameters = [{}]
        if self.dump:
            self.logger.info("Output directory: %s", self.outdir)

        return parameters

    def _run_iters_for_params(
        self,
        params
    ):
        """Run the simulation and yield the resulting environments."""

        with serialization.with_source(self.source_file):
            with utils.timer(f"running for config {params}"):
                if self.dry_run:
                    def func(*args, **kwargs):
                        return None
                else:
                    func = self._run_model

                for env in tqdm(utils.run_parallel(
                    func=func,
                    iterable=range(self.iterations),
                    **params,
                ), total=self.iterations, leave=False):
                    if env is None and self.dry_run:
                        continue

                    yield env

    def _get_env(self, iteration_id, params):
        """Create an environment for a iteration of the simulation"""

        iteration_id = str(iteration_id)

        agent_reporters = self.agent_reporters
        agent_reporters.update(params.pop("agent_reporters", {}))
        model_reporters = self.model_reporters
        model_reporters.update(params.pop("model_reporters", {}))

        return self.model(
            id=iteration_id,
            seed=f"{self.seed}_iteration_{iteration_id}",
            dir_path=self.dir_path,
            logger=self.logger.getChild(iteration_id),
            agent_reporters=agent_reporters,
            model_reporters=model_reporters,
            tables=self.tables,
            **params,
        )

    def _run_model(self, iteration_id, **params):
        """
        Run a single iteration of the simulation

        """
        # Set-up iteration environment and graph
        model = self._get_env(iteration_id, params)
        with utils.timer("Simulation {} iteration {}".format(self.name, iteration_id)):

            max_time = self.max_time 
            max_steps = self.max_steps

            if (max_time is not None) and (max_steps is not None):
                is_done = lambda model: (not model.running) or (model.schedule.time >= max_time) or (model.schedule.steps >= max_steps)
            elif max_time is not None:
                is_done = lambda model: (not model.running) or (model.schedule.time >= max_time)
            elif max_steps is not None:
                is_done = lambda model: (not model.running) or (model.schedule.steps >= max_steps)
            else:
                is_done = lambda model: not model.running
            
            if not model.schedule.agents:
                raise Exception("No agents in model. This is probably a bug. Make sure that the model has agents scheduled after its initialization.")

            newline = "\n"
            self.logger.debug(
                dedent(
                    f"""
    Model stats:
    Agent count: { model.schedule.get_agent_count() }):
    Topology size: { len(model.G) if hasattr(model, "G") else 0 }
            """
                )
            )

            if self.debug:
                set_trace()

            while not is_done(model):
                self.logger.debug(
                    f'Simulation time {model.schedule.time}/{max_time}.'
                )
                model.step()

        return model

    def to_dict(self):
        d = asdict(self)
        return serialization.serialize_dict(d)

    def to_yaml(self):
        return yaml.dump(self.to_dict())


def iter_from_file(*files, **kwargs):
    for f in files:
        try:
            yield from iter_from_py(f, **kwargs)
        except ValueError as ex:
            yield from iter_from_config(f, **kwargs)


def from_file(*args, **kwargs):
    return list(iter_from_file(*args, **kwargs))


def iter_from_config(*cfgs, **kwargs):
    for config in cfgs:
        configs = list(serialization.load_config(config))
        for config, path in configs:
            d = dict(config)
            d.update(kwargs)
            if "dir_path" not in d:
                d["dir_path"] = os.path.dirname(path)
            yield Simulation(**d)


def from_config(conf_or_path):
    lst = list(iter_from_config(conf_or_path))
    if len(lst) > 1:
        raise AttributeError("Provide only one configuration")
    return lst[0]


def iter_from_py(pyfile, module_name='imported_file', **kwargs):
    """Try to load every Simulation instance in a given Python file"""
    import importlib
    added = False
    sims = []
    assert not _AVOID_RUNNING
    with do_not_run():
        assert _AVOID_RUNNING
        spec = importlib.util.spec_from_file_location(module_name, pyfile)
        folder = os.path.dirname(pyfile)
        if folder not in sys.path:
            added = True
            sys.path.append(folder)
        if not spec:
            raise ValueError(f"{pyfile} does not seem to be a Python module")
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)
        for (_name, sim) in inspect.getmembers(module, lambda x: isinstance(x, Simulation)):
            sims.append(sim)
        for sim in _iter_queued():
            sims.append(sim)
        if not sims:
            for (_name, env) in inspect.getmembers(module,
                                                   lambda x: inspect.isclass(x) and
                                                   issubclass(x, environment.Environment) and
                                                   (getattr(x, "__module__", None) != environment.__name__)):
                sims.append(Simulation(model=env, **kwargs))
        del sys.modules[module_name]
    assert not _AVOID_RUNNING
    if not sims:
        raise AttributeError(f"No valid configurations found in {pyfile}")
    if added:
        sys.path.remove(folder)
    for sim in sims:
        yield replace(sim, **kwargs)


def from_py(pyfile):
    return next(iter_from_py(pyfile))


def run_from_file(*files, **kwargs):
    for sim in iter_from_file(*files):
        logger.info(f"Using config(s): {sim.name}")
        sim.run_simulation(**kwargs)

def run(env, iterations=1, num_processes=1, dump=False, name="test", **kwargs):
    return Simulation(model=env, iterations=iterations, name=name, dump=dump, num_processes=num_processes, **kwargs).run()


from . import BaseAgent


class NetworkAgent(BaseAgent):
    def __init__(self, *args, topology=None, init=True, node_id=None, **kwargs):
        super().__init__(*args, init=False, **kwargs)

        self.G = topology or self.model.G
        assert self.G is not None, "Network agents should have a network"
        if node_id is None:
            nodes = self.random.choices(list(self.G.nodes), k=len(self.G))
            for n_id in nodes:
                if "agent" not in self.G.nodes[n_id] or self.G.nodes[n_id]["agent"] is None:
                    node_id = n_id
                    break
            else:
                node_id = len(self.G)
                self.info(f"All nodes ({len(self.G)}) have an agent assigned, adding a new node to the graph for agent {self.unique_id}")
                self.G.add_node(node_id)
        assert node_id is not None
        self.G.nodes[node_id]["agent"] = self
        self.node_id = node_id
        if init:
            self.init()

    def count_neighbors(self, state_id=None, **kwargs):
        return len(self.get_neighbors(state_id=state_id, **kwargs))

    def iter_neighbors(self, **kwargs):
        return self.iter_agents(limit_neighbors=True, **kwargs)

    def get_neighbors(self, **kwargs):
        return list(self.iter_neighbors(**kwargs))

    @property
    def node(self):
        return self.G.nodes[self.node_id]

    def iter_agents(self, unique_id=None, *, limit_neighbors=False, **kwargs):
        unique_ids = None
        if unique_ids is not None:
            try:
                unique_ids = set(unique_id)
            except TypeError:
                unique_ids = set([unique_id])

        if limit_neighbors:
            neighbor_ids = set()
            for node_id in self.G.neighbors(self.node_id):
                agent = self.G.nodes[node_id].get("agent")
                if agent is not None:
                    neighbor_ids.add(agent.unique_id)
            if unique_ids:
                unique_ids = unique_ids & neighbor_ids
            else:
                unique_ids = neighbor_ids
            if not unique_ids:
                return
            unique_ids = list(unique_ids)
        yield from super().iter_agents(unique_id=unique_ids, **kwargs)

    def subgraph(self, center=True, **kwargs):
        include = [self] if center else []
        G = self.G.subgraph(
            n.node_id for n in list(self.get_agents(**kwargs) + include)
        )
        return G

    def remove_node(self):
        self.debug(f"Removing node for {self.unique_id}: {self.node_id}")
        self.G.remove_node(self.node_id)
        self.node_id = None

    def add_edge(self, other, edge_attr_dict=None, *edge_attrs):
        if self.node_id not in self.G.nodes(data=False):
            raise ValueError(
                "{} not in list of existing agents in the network".format(
                    self.unique_id
                )
            )
        if other.node_id not in self.G.nodes(data=False):
            raise ValueError(
                "{} not in list of existing agents in the network".format(other)
            )

        self.G.add_edge(
            self.node_id, other.node_id, edge_attr_dict=edge_attr_dict, *edge_attrs
        )

    def die(self, remove=True):
        if not self.alive:
            return None
        if remove:
            self.remove_node()
        return super().die()


NetAgent = NetworkAgent



from mesa.time import BaseScheduler
from queue import Empty
from heapq import heappush, heappop, heapreplace
from collections import deque
import math
import logging

from inspect import getsource
from numbers import Number
from textwrap import dedent

from .utils import logger
from mesa import Agent as MesaAgent


INFINITY = float("inf")

class Delay:
    """A delay object which can be used both as a return value and as an awaitable (in async code)."""
    __slots__ = ("delta", )
    def __init__(self, delta):
        self.delta = float(delta)

    def __float__(self):
        return self.delta
    
    def __await__(self):
        return (yield self.delta)


class DeadAgent(Exception):
    pass


class DiscreteActivation(BaseScheduler):
    default_interval = 1
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if hasattr(self.model, 'default_interval'):
            self.default_interval = self.model.interval


class PQueueActivation(DiscreteActivation):
    """
    A scheduler which activates each agent with a delay returned by the agent's step method.
    If no delay is returned, a default of 1 is used.
    
    In each activation, each agent will update its 'next_time'.
    """

    def __init__(self, *args, shuffle=True, **kwargs):
        super().__init__(*args, **kwargs)
        self._queue = []
        self._shuffle = shuffle
        self.logger = getattr(self.model, "logger", logger).getChild(f"time_{ self.model }")
        self.next_time = self.time

    def add(self, agent: MesaAgent, when=None):
        if when is None:
            when = self.time
        else:
            when = float(when)

        self._schedule(agent, None, when)
        super().add(agent)

    def _schedule(self, agent, when=None, replace=False):
        if when is None:
            when = self.time
        if self._shuffle:
            key = (when, self.model.random.random())
        else:
            key = (when, agent.unique_id)
        if replace:
            heapreplace(self._queue, (key, agent))
        else:
            heappush(self._queue, (key, agent))

    def step(self) -> None:
        """
        Executes agents in order, one at a time. After each step,
        an agent will signal when it wants to be scheduled next.
        """

        if self.time == INFINITY:
            return

        next_time = INFINITY

        now = self.time

        while self._queue:
            ((when, _id), agent) = self._queue[0]
            if when > now:
                next_time = when
                break

            try:
                when = agent.step() or self.default_interval
                when += now
            except DeadAgent:
                heappop(self._queue)
                continue

            if when == INFINITY:
                heappop(self._queue)
                continue

            self._schedule(agent, when, replace=True)

        self.steps += 1

        self.time = next_time

        if next_time == INFINITY:
            self.model.running = False
            self.time = INFINITY
            return


class TimedActivation(DiscreteActivation):
    '''A discrete-time scheduler that has time buckets with agents that should be woken at the same time instant.'''
    def __init__(self, *args, shuffle=True, **kwargs):
        super().__init__(*args, **kwargs)
        self._queue = deque()
        self._shuffle = shuffle
        self.logger = getattr(self.model, "logger", logger).getChild(f"time_{ self.model }")
        self.next_time = self.time

    def add(self, agent: MesaAgent, when=None):
        if when is None:
            when = self.time
        else:
            when = float(when)
        self._schedule(agent, None, when)
        super().add(agent)

    def _schedule(self, agent, when=None, replace=False):
        when = when or self.time
        pos = len(self._queue)
        for (ix, l) in enumerate(self._queue):
            if l[0] == when:
                l[1].append(agent)
                return
            if l[0] > when:
                pos = ix
                break
        self._queue.insert(pos, (when, [agent]))

    def step(self) -> None:
        """
        Executes agents in order, one at a time. After each step,
        an agent will signal when it wants to be scheduled next.
        """
        if not self._queue:
            return

        now = self.time

        next_time = self._queue[0][0]

        if next_time > now:
            self.time = next_time
            return

        bucket = self._queue.popleft()[1]
        if self._shuffle:
            self.model.random.shuffle(bucket)
        for agent in bucket:
            try:
                when = agent.step() or self.default_interval
                when += now
            except DeadAgent:
                continue

            if when != INFINITY:
                self._schedule(agent, when, replace=True)

        self.steps += 1
        if self._queue:
            self.time = self._queue[0][0]
        else:
            self.time = INFINITY


class ShuffledTimedActivation(TimedActivation):
    '''
    A TimedActivation scheduler that processes events in random order.
    '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, shuffle=True, **kwargs)


class OrderedTimedActivation(TimedActivation):
    '''
    A TimedActivation scheduler that always processes events in
    the same order.
    '''
    def __init__(self, *args, **kwargs):
        super().__init__(*args, shuffle=False, **kwargs)


Scheduler = TimedActivation

    
class Lockstepper:
    '''
    A wrapper class to produce discrete-event schedulers that behave like
    fixed-time schedulers.
    '''
    
    def  __init__(self, scheduler: BaseScheduler, interval=1):
        self.scheduler = scheduler
        self.default_interval = interval
        self.time = scheduler.time
        self.steps = 0

    def step(self):
        end_time = self.time + self.default_interval
        res = None
        while self.scheduler.time < end_time:
            res = self.scheduler.step()
        self.time = end_time
        self.steps += 1
        return res

    def __getattr__(self, name):
        return getattr(self.scheduler, name)


from dataclasses import dataclass, field
from typing import Any
from uuid import uuid4


class Event:
    pass


@dataclass
class Message:
    payload: Any
    sender: Any = None
    expiration: float = None
    timestamp: float = None
    id: int = field(default_factory=uuid4)

    def expired(self, when):
        return self.expiration is not None and self.expiration < when


class Reply(Message):
    source: Message


class Ask(Message):
    reply: Message = None


class Tell(Message):
    pass


class TimedOut(Exception):
    pass



from . import MetaAgent, BaseAgent
from .. import time
from types import coroutine
from functools import partial, wraps
import inspect


class State:
    __slots__ = ("awaitable", "f", "generator", "name", "default")

    def __init__(self, f, name, default, generator, awaitable):
        self.f = f
        self.name = name
        self.generator = generator
        self.awaitable = awaitable
        self.default = default

    @coroutine
    def step(self, obj):
        if self.generator or self.awaitable:
            f = self.f
            next_state = yield from f(obj)
            return next_state

        else:
            return self.f(obj)

    @property
    def id(self):
        return self.name

    def __call__(self, *args, **kwargs):
        raise Exception("States should not be called directly")

class UnboundState(State):

    def bind(self, obj):
        bs = BoundState(self.f, self.name, self.default, self.generator, self.awaitable, obj=obj)
        setattr(obj, self.name, bs)
        return bs


class BoundState(State):
    __slots__ = ("obj", )

    def __init__(self, *args, obj):
        super().__init__(*args)
        self.obj = obj
    
    def delay(self, delta=0):
        return self, self.obj.delay(delta)
    
    def at(self, when):
        return self, self.obj.at(when)


def state(name=None, default=False):
    def decorator(func, name=None):
        """
        A state function should return either a state id, or a tuple (state_id, when)
        The default value for state_id is the current state id.
        """
        name = name or func.__name__
        generator = inspect.isgeneratorfunction(func)
        awaitable = inspect.iscoroutinefunction(func) or inspect.isasyncgen(func)
        return UnboundState(func, name, default, generator, awaitable)

    if callable(name):
        return decorator(name)
    else:
        return partial(decorator, name=name)


def default_state(func):
    func.default = True
    return func


class MetaFSM(MetaAgent):
    def __new__(mcls, name, bases, namespace):
        states = {}
        # Re-use states from inherited classes
        default_state = None
        for i in bases:
            if isinstance(i, MetaFSM):
                for state_id, state in i._states.items():
                    if state.default:
                        default_state = state
                    states[state_id] = state

        # Add new states
        for attr, func in namespace.items():
            if isinstance(func, State):
                if func.default:
                    default_state = func
                states[func.name] = func

        namespace.update(
            {
                "_state": default_state,
                "_states": states,
            }
        )

        cls = super(MetaFSM, mcls).__new__(
            mcls=mcls, name=name, bases=bases, namespace=namespace
        )
        for (k, v) in states.items():
            setattr(cls, k, v)
        return cls


class FSM(BaseAgent, metaclass=MetaFSM):
    def __init__(self, init=True, state_id=None, **kwargs):
        super().__init__(**kwargs, init=False)
        if state_id is not None:
            self._set_state(state_id)
        # If more than "dead" state is defined, but no default state
        if len(self._states) > 1 and not self._state:
            raise ValueError(
                f"No default state specified for {type(self)}({self.unique_id})"
            )
        for (k, v) in self._states.items():
            setattr(self, k, v.bind(self))

        if init:
            self.init()

    @classmethod
    def states(cls):
        return list(cls._states.keys())

    @property
    def state_id(self):
        return self._state.name
    
    def set_state(self, value):
        if self.now > 0:
            raise ValueError("Cannot change state after init")
        self._set_state(value)

    def step(self):
        self._check_alive()
        next_state = yield from self._state.step(self)

        try:
            next_state, when = next_state
        except (TypeError, ValueError) as ex:
            try:
                self._set_state(next_state)
                return None
            except ValueError:
                return next_state

        self._set_state(next_state)
        return when

    def _set_state(self, state):
        if state is None:
            return
        if isinstance(state, str):
            if state not in self._states:
                raise ValueError("{} is not a valid state".format(state))
            state = self._states[state]
        if not isinstance(state, State):
            raise ValueError("{} is not a valid state".format(state))
        self._state = state

    def die(self, *args, **kwargs):
        super().die(*args, **kwargs)
        return self.dead.at(time.INFINITY)

    @state
    def dead(self):
        return time.INFINITY


from __future__ import annotations

import logging
from collections import OrderedDict, defaultdict
from collections.abc import MutableMapping, Mapping, Set
from abc import ABCMeta
from copy import deepcopy, copy
from functools import partial, wraps
from itertools import islice, chain
import inspect
import types
import textwrap
import networkx as nx
import warnings
import sys

from typing import Any

from mesa import Agent as MesaAgent, Model
from typing import Dict, List

from .. import serialization, network, utils, time, config


IGNORED_FIELDS = ("model", "logger")


def decorate_generator_step(func, name):
    @wraps(func)
    def decorated(self):
        while True:
            if self._coroutine is None:
                self._coroutine = func(self)
            try:
                if self._last_except:
                    val = self._coroutine.throw(self._last_except)
                else:
                    val = self._coroutine.send(self._last_return)
            except StopIteration as ex:
                self._coroutine = None
                val = ex.value
            finally:
                self._last_return = None
                self._last_except = None
            return float(val) if val is not None else val
    return decorated


def decorate_normal_func(func, name):
    @wraps(func)
    def decorated(self):
        val = func(self)
        return float(val) if val is not None else val
    return decorated


class MetaAgent(ABCMeta):
    def __new__(mcls, name, bases, namespace):
        defaults = {}

        # Re-use defaults from inherited classes
        for i in bases:
            if isinstance(i, MetaAgent):
                defaults.update(i._defaults)

        new_nmspc = {
            "_defaults": defaults,
        }

        for attr, func in namespace.items():
            if attr == "step":
                if inspect.isgeneratorfunction(func) or inspect.iscoroutinefunction(func):
                    func = decorate_generator_step(func, attr)
                    new_nmspc.update({
                        "_last_return": None,
                        "_last_except": None,
                        "_coroutine": None,
                    })
                elif inspect.isasyncgenfunction(func):
                    raise ValueError("Illegal step function: {}. It probably mixes both async/await and yield".format(func))
                elif inspect.isfunction(func):
                    func = decorate_normal_func(func, attr)
                else:
                    raise ValueError("Illegal step function: {}".format(func))
                new_nmspc[attr] = func
            elif (
                isinstance(func, types.FunctionType)
                or isinstance(func, property)
                or isinstance(func, classmethod)
                or attr[0] == "_"
            ):
                new_nmspc[attr] = func
            elif attr == "defaults":
                defaults.update(func)
            elif inspect.isfunction(func):
                new_nmspc[attr] = func
            else:
                defaults[attr] = copy(func)


        # Add attributes for their use in the decorated functions
        return super().__new__(mcls, name, bases, new_nmspc)


class BaseAgent(MesaAgent, MutableMapping, metaclass=MetaAgent):
    """
    A special type of Mesa Agent that:

    * Can be used as a dictionary to access its state.
    * Has logging built-in
    * Can be given default arguments through a defaults class attribute,
    which will be used on construction to initialize each agent's state

    Any attribute that is not preceded by an underscore (`_`) will also be added to its state.
    """

    def __init__(self, unique_id, model, name=None, init=True, **kwargs):
        assert isinstance(unique_id, int)
        super().__init__(unique_id=unique_id, model=model)

        self.name = (
            str(name) if name else "{}[{}]".format(type(self).__name__, self.unique_id)
        )

        self.alive = True

        logger = model.logger.getChild(self.name)
        self.logger = logging.LoggerAdapter(logger, {"agent_name": self.name})

        if hasattr(self, "level"):
            self.logger.setLevel(self.level)

        for k in self._defaults:
            v = getattr(model, k, None)
            if v is not None:
                setattr(self, k, v)

        for (k, v) in self._defaults.items():
            if not hasattr(self, k) or getattr(self, k) is None:
                setattr(self, k, deepcopy(v))

        for (k, v) in kwargs.items():
            setattr(self, k, v)

        if init:
            self.init()

    def init(self):
        pass

    def __hash__(self):
        return hash(self.unique_id)

    def prob(self, probability):
        return prob(probability, self.model.random)

    @classmethod
    def w(cls, **kwargs):
        return custom(cls, **kwargs)

    # TODO: refactor to clean up mesa compatibility
    @property
    def id(self):
        msg = "This attribute is deprecated. Use `unique_id` instead"
        warnings.warn(msg, DeprecationWarning)
        print(msg, file=sys.stderr)
        return self.unique_id

    @classmethod
    def from_dict(cls, model, attrs, warn_extra=True):
        ignored = {}
        args = {}
        for k, v in attrs.items():
            if k in inspect.signature(cls).parameters:
                args[k] = v
            else:
                ignored[k] = v
        if ignored and warn_extra:
            utils.logger.info(
                f"Ignoring the following arguments for agent class { agent_class.__name__ }: { ignored }"
            )
        return cls(model=model, **args)

    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(f"key {key}  not found in agent")

    def __delitem__(self, key):
        return delattr(self, key)

    def __contains__(self, key):
        return hasattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __len__(self):
        return sum(1 for n in self.keys())

    def __iter__(self):
        return self.items()

    def keys(self):
        return (k for k in self.__dict__ if k[0] != "_" and k not in IGNORED_FIELDS)

    def items(self, keys=None, skip=None):
        keys = keys if keys is not None else self.keys()
        it = ((k, self.get(k, None)) for k in keys)
        if skip:
            return filter(lambda x: x[0] not in skip, it)
        return it

    def get(self, key, default=None):
        try:
            return getattr(self, key)
        except AttributeError:
            try:
                return getattr(self.model, key)
            except AttributeError:
                return default

    @property
    def now(self):
        try:
            return self.model.now
        except AttributeError:
            # No environment
            return None

    def die(self, msg=None):
        if msg:
            self.debug("Agent dying:", msg)
        else:
            self.debug(f"agent dying")
        self.alive = False
        try:
            self.model.schedule.remove(self)
        except KeyError:
            pass
        return time.Delay(time.INFINITY)

    def step(self):
        raise NotImplementedError("Agent must implement step method")

    def _check_alive(self):
        if not self.alive:
            raise time.DeadAgent(self.unique_id)

    def log(self, *message, level=logging.INFO, **kwargs):
        if not self.logger.isEnabledFor(level):
            return
        message = " ".join(str(i) for i in message)
        message = "[@{:>4}]\t{:>10}: {}".format(self.now, repr(self), message)
        for k, v in kwargs:
            message += " {k}={v} ".format(k, v)
        extra = {}
        extra["now"] = self.now
        extra["unique_id"] = self.unique_id
        extra["agent_name"] = self.name
        return self.logger.log(level, message, extra=extra)

    def debug(self, *args, **kwargs):
        return self.log(*args, level=logging.DEBUG, **kwargs)

    def info(self, *args, **kwargs):
        return self.log(*args, level=logging.INFO, **kwargs)

    def count_agents(self, **kwargs):
        return len(list(self.get_agents(**kwargs)))

    def get_agents(self, *args, **kwargs):
        it = self.iter_agents(*args, **kwargs)
        return list(it)

    def iter_agents(self, *args, **kwargs):
        yield from filter_agents(self.model.schedule._agents, *args, **kwargs)

    def __str__(self):
        return self.to_str()

    def to_str(self, keys=None, skip=None, pretty=False):
        content = dict(self.items(keys=keys))
        if pretty and content:
            d = content
            content = "\n"
            for k, v in d.items():
                content += f"- {k}: {v}\n"
            content = textwrap.indent(content, "    ")
        return f"{repr(self)}{content}"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.unique_id})"
    
    def at(self, at):
        return time.Delay(float(at) - self.now)
    
    def delay(self, delay=1):
        return time.Delay(delay)


def prob(prob, random):
    """
    A true/False uniform distribution with a given probability.
    To be used like this:

    .. code-block:: python

          if prob(0.3):
              do_something()

    """
    r = random.random()
    return r < prob


def calculate_distribution(network_agents=None, agent_class=None):
    """
    Calculate the threshold values (thresholds for a uniform distribution)
    of an agent distribution given the weights of each agent type.

    The input has this form: ::

            [
            {'agent_class': 'agent_class_1',
                'weight': 0.2,
                'state': {
                    'id': 0
                }
            },
            {'agent_class': 'agent_class_2',
                'weight': 0.8,
                'state': {
                    'id': 1
                }
            }
            ]

    In this example, 20% of the nodes will be marked as type
    'agent_class_1'.
    """
    if network_agents:
        network_agents = [
            deepcopy(agent) for agent in network_agents if not hasattr(agent, "id")
        ]
    elif agent_class:
        network_agents = [{"agent_class": agent_class}]
    else:
        raise ValueError("Specify a distribution or a default agent type")

    # Fix missing weights and incompatible types
    for x in network_agents:
        x["weight"] = float(x.get("weight", 1))

    # Calculate the thresholds
    total = sum(x["weight"] for x in network_agents)
    acc = 0
    for v in network_agents:
        if "ids" in v:
            continue
        upper = acc + (v["weight"] / total)
        v["threshold"] = [acc, upper]
        acc = upper
    return network_agents


def _serialize_type(agent_class, known_modules=[], **kwargs):
    if isinstance(agent_class, str):
        return agent_class
    known_modules += ["soil.agents"]
    return serialization.serialize(agent_class, known_modules=known_modules, **kwargs)[
        1
    ]  # Get the name of the class


def _deserialize_type(agent_class, known_modules=[]):
    if not isinstance(agent_class, str):
        return agent_class
    known = known_modules + ["soil.agents", "soil.agents.custom"]
    agent_class = serialization.deserializer(agent_class, known_modules=known)
    return agent_class


class AgentView(Mapping, Set):
    """A lazy-loaded list of agents."""

    __slots__ = ("_agents",)

    def __init__(self, agents):
        self._agents = agents

    def __getstate__(self):
        return {"_agents": self._agents}

    def __setstate__(self, state):
        self._agents = state["_agents"]

    # Mapping methods
    def __len__(self):
        return len(self._agents)

    def __iter__(self):
        yield from self._agents.values()

    def __getitem__(self, agent_id):
        if isinstance(agent_id, slice):
            raise ValueError(f"Slicing is not supported")
        if agent_id in self._agents:
            return self._agents[agent_id]
        raise ValueError(f"Agent {agent_id} not found")

    def filter(self, *args, **kwargs):
        yield from filter_agents(self._agents, *args, **kwargs)

    def one(self, *args, **kwargs):
        return next(filter_agents(self._agents, *args, **kwargs))

    def __call__(self, *args, **kwargs):
        return list(self.filter(*args, **kwargs))

    def __contains__(self, agent_id):
        return agent_id in self._agents

    def __str__(self):
        return str(list(unique_id for unique_id in self.keys()))

    def __repr__(self):
        return f"{self.__class__.__name__}({self})"


def filter_agents(
    agents: dict,
    *id_args,
    unique_id=None,
    state_id=None,
    agent_class=None,
    state=None,
    limit=None,
    **kwargs,
):
    """
    Filter agents given as a dict, by the criteria given as arguments (e.g., certain type or state id).
    """

    ids = []

    if unique_id is not None:
        if isinstance(unique_id, list):
            ids += unique_id
        else:
            ids.append(unique_id)

    if id_args:
        ids += id_args

    if ids:
        f = (agent for agent in agents if agent.unique_id in ids)
    else:
        f = agents

    if state_id is not None and not isinstance(state_id, (tuple, list)):
        state_id = tuple([state_id])

    if agent_class is not None:
        agent_class = _deserialize_type(agent_class)
        try:
            agent_class = tuple(agent_class)
        except TypeError:
            agent_class = tuple([agent_class])

    if state_id is not None:
        f = filter(lambda agent: agent.get("state_id", None) in state_id, f)

    if agent_class is not None:
        f = filter(lambda agent: isinstance(agent, agent_class), f)

    state = state or dict()
    state.update(kwargs)

    for k, vs in state.items():
        if not isinstance(vs, list):
            vs = [vs]
        f = filter(lambda agent: any(getattr(agent, k, None) == v for v in vs), f)

    if limit is not None:
        f = islice(f, limit)

    yield from f


def from_config(
    cfg: config.AgentConfig, random, topology: nx.Graph = None
) -> List[Dict[str, Any]]:
    """
    This function turns an agentconfig into a list of individual "agent specifications", which are just a dictionary
    with the parameters that the environment will use to construct each agent.

    This function does NOT return a list of agents, mostly because some attributes to the agent are not known at the
    time of calling this function, such as `unique_id`.
    """
    default = cfg or config.AgentConfig()
    if not isinstance(cfg, config.AgentConfig):
        cfg = config.AgentConfig(**cfg)

    agents = []

    assigned_total = 0
    assigned_network = 0

    if cfg.fixed is not None:
        agents, assigned_total, assigned_network = _from_fixed(
            cfg.fixed, topology=cfg.topology, default=cfg
        )

    n = cfg.n

    if cfg.distribution:
        topo_size = len(topology) if topology else 0

        networked = []
        total = []

        for d in cfg.distribution:
            if d.strategy == config.Strategy.topology:
                topo = d.topology if ("topology" in d.__fields_set__) else cfg.topology
                if not topo:
                    raise ValueError(
                        'The "topology" strategy only works if the topology parameter is set to True'
                    )
                if not topo_size:
                    raise ValueError(
                        f"Topology does not have enough free nodes to assign one to the agent"
                    )

                networked.append(d)

            if d.strategy == config.Strategy.total:
                if not cfg.n:
                    raise ValueError(
                        'Cannot use the "total" strategy without providing the total number of agents'
                    )
                total.append(d)

        if networked:
            new_agents = _from_distro(
                networked,
                n=topo_size - assigned_network,
                topology=topo,
                default=cfg,
                random=random,
            )
            assigned_total += len(new_agents)
            assigned_network += len(new_agents)
            agents += new_agents

        if total:
            remaining = n - assigned_total
            agents += _from_distro(total, n=remaining, default=cfg, random=random)

        if assigned_network < topo_size:
            utils.logger.warn(
                f"The total number of agents does not match the total number of nodes in "
                "every topology. This may be due to a definition error: assigned: "
                f"{ assigned } total size: { topo_size }"
            )

    return agents


def _from_fixed(
    lst: List[config.FixedAgentConfig],
    topology: bool,
    default: config.SingleAgentConfig,
) -> List[Dict[str, Any]]:
    agents = []

    counts_total = 0
    counts_network = 0

    for fixed in lst:
        agent = {}
        if default:
            agent = default.state.copy()
        agent.update(fixed.state)
        cls = serialization.deserialize(
            fixed.agent_class or (default and default.agent_class)
        )
        agent["agent_class"] = cls
        topo = (
            fixed.topology
            if ("topology" in fixed.__fields_set__)
            else topology or default.topology
        )

        if topo:
            agent["topology"] = True
            counts_network += 1
        if not fixed.hidden:
            counts_total += 1
        agents.append(agent)

    return agents, counts_total, counts_network


def _from_distro(
    distro: List[config.AgentDistro],
    n: int,
    default: config.SingleAgentConfig,
    random,
    topology: str = None
) -> List[Dict[str, Any]]:

    agents = []

    if n is None:
        if any(lambda dist: dist.n is None, distro):
            raise ValueError(
                "You must provide a total number of agents, or the number of each type"
            )
        n = sum(dist.n for dist in distro)

    weights = list(dist.weight if dist.weight is not None else 1 for dist in distro)
    minw = min(weights)
    norm = list(weight / minw for weight in weights)
    total = sum(norm)
    chunk = n // total

    # random.choices would be enough to get a weighted distribution. But it can vary a lot for smaller k
    # So instead we calculate our own distribution to make sure the actual ratios are close to what we would expect

    # Calculate how many times each has to appear
    indices = list(
        chain.from_iterable([idx] * int(n * chunk) for (idx, n) in enumerate(norm))
    )

    # Complete with random agents following the original weight distribution
    if len(indices) < n:
        indices += random.choices(
            list(range(len(distro))),
            weights=[d.weight for d in distro],
            k=n - len(indices),
        )

    # Deserialize classes for efficiency
    classes = list(
        serialization.deserialize(i.agent_class or default.agent_class) for i in distro
    )

    # Add them in random order
    random.shuffle(indices)

    for idx in indices:
        d = distro[idx]
        agent = d.state.copy()
        cls = classes[idx]
        agent["agent_class"] = cls
        if default:
            agent.update(default.state)
        topology = (
            d.topology
            if ("topology" in d.__fields_set__)
            else topology or default.topology
        )
        if topology:
            agent["topology"] = topology
        agents.append(agent)

    return agents


from .network_agents import *
from .fsm import *
from .evented import *
from typing import Optional


class Agent(NetworkAgent, FSM, EventedAgent):
    """Default agent class, has both network and event capabilities"""


from ..environment import NetworkEnvironment


from .BassModel import *
from .IndependentCascadeModel import *
from .SISaModel import *
from .CounterModel import *


def custom(cls, **kwargs):
    """Create a new class from a template class and keyword arguments"""
    return type(cls.__name__, (cls,), kwargs)



from __future__ import annotations

import pdb
import sys
import os

from textwrap import indent
from functools import wraps

from .agents import FSM, MetaFSM
from mesa import Model, Agent


def wrapcmd(func):
    @wraps(func)
    def wrapper(self, arg: str, temporary=False):
        sys.settrace(self.trace_dispatch)

        lastself = self
        known = globals()
        known.update(self.curframe.f_globals)
        known.update(self.curframe.f_locals)
        known["attrs"] = arg.strip().split()

        this = known.get("self", None)

        if isinstance(this, Model):
            known["model"] = this
        elif isinstance(this, Agent):
            known["agent"] = this
            known["model"] = this.model

        known["self"] = lastself
        return exec(func.__code__, known, known)

    return wrapper


class Debug(pdb.Pdb):
    def __init__(self, *args, skip_soil=False, **kwargs):
        skip = kwargs.get("skip", [])
        if skip_soil:
            skip.append("soil")
            skip.append("contextlib")
            skip.append("soil.*")
            skip.append("mesa.*")
        super(Debug, self).__init__(*args, skip=skip, **kwargs)
        self.prompt = "[soil-pdb] "

    @staticmethod
    def _soil_agents(model, attrs=None, pretty=True, **kwargs):
        for agent in model.get_agents(**kwargs):
            d = agent
            print(" - " + indent(agent.to_str(keys=attrs, pretty=pretty), "  "))

    @wrapcmd
    def do_soil_agents():
        return Debug._soil_agents(model, attrs=attrs or None)

    do_sa = do_soil_agents

    @wrapcmd
    def do_soil_list():
        return Debug._soil_agents(model, attrs=["state_id"], pretty=False)

    do_sl = do_soil_list

    def do_continue_state(self, arg):
        """Continue until next time this state is reached"""
        self.do_break_state(arg, temporary=True)
        return self.do_continue("")

    do_cs = do_continue_state

    @wrapcmd
    def do_soil_agent():
        if not agent:
            print("No agent available")
            return

        keys = None
        if attrs:
            keys = []
            for k in attrs:
                for key in agent.keys():
                    if key.startswith(k):
                        keys.append(key)

        print(agent.to_str(pretty=True, keys=keys))

    do_aa = do_soil_agent

    def do_break_step(self, arg: str):
        """
        Break before the next step.
        """
        try:
            known = globals()
            known.update(self.curframe.f_globals)
            known.update(self.curframe.f_locals)
            func = getattr(known["model"], "step")
        except AttributeError as ex:
            self.error(f"The model does not have a step function: {ex}")
            return
        if hasattr(func, "__func__"):
            func = func.__func__

        code = func.__code__
        # use co_name to identify the bkpt (function names
        # could be aliased, but co_name is invariant)
        funcname = code.co_name
        lineno = code.co_firstlineno
        filename = code.co_filename

        # Check for reasonable breakpoint
        line = self.checkline(filename, lineno)
        if not line:
            raise ValueError("no line found")
            # now set the break point

        existing = self.get_breaks(filename, line)
        if existing:
            self.message("Breakpoint already exists at %s:%d" % (filename, line))
            return
        cond = f"self.schedule.steps > {model.schedule.steps}"
        err = self.set_break(filename, line, True, cond, funcname)
        if err:
            self.error(err)
        else:
            bp = self.get_breaks(filename, line)[-1]
            self.message("Breakpoint %d at %s:%d" % (bp.number, bp.file, bp.line))
            return self.do_continue("")
    
    do_bstep = do_break_step

    def do_break_state(self, arg: str, instances=None, temporary=False):
        """
        Break before a specified state is stepped into.
        """

        klass = None
        state = arg
        if not state:
            self.error("Specify at least a state name")
            return

        state, *tokens = state.lstrip().split()
        if tokens:
            instances = list(eval(token) for token in tokens)

        colon = state.find(":")

        if colon > 0:
            klass = state[:colon].rstrip()
            state = state[colon + 1 :].strip()

            print(klass, state, tokens)
            klass = eval(klass, self.curframe.f_globals, self.curframe_locals)

        if klass:
            klasses = [klass]
        else:
            klasses = [
                k
                for k in self.curframe.f_globals.values()
                if isinstance(k, type) and issubclass(k, FSM)
            ]

        if not klasses:
            self.error("No agent classes found")

        for klass in klasses:
            try:
                func = getattr(klass, state)
            except AttributeError:
                self.error(f"State {state} not found in class {klass}")
                continue
            if hasattr(func, "__func__"):
                func = func.__func__

            code = func.__code__
            # use co_name to identify the bkpt (function names
            # could be aliased, but co_name is invariant)
            funcname = code.co_name
            lineno = code.co_firstlineno
            filename = code.co_filename

            # Check for reasonable breakpoint
            line = self.checkline(filename, lineno)
            if not line:
                raise ValueError("no line found")
                # now set the break point
            cond = None
            if instances:
                cond = f"self.unique_id in { repr(instances) }"

            existing = self.get_breaks(filename, line)
            if existing:
                self.message("Breakpoint already exists at %s:%d" % (filename, line))
                continue
            err = self.set_break(filename, line, temporary, cond, funcname)
            if err:
                self.error(err)
            else:
                bp = self.get_breaks(filename, line)[-1]
                self.message("Breakpoint %d at %s:%d" % (bp.number, bp.file, bp.line))

    do_bs = do_break_state

    def do_break_state_self(self, arg: str, temporary=False):
        """
        Break before a specified state is stepped into, for the current agent
        """
        agent = self.curframe.f_locals.get("self")
        if not agent:
            self.error("No current agent.")
            self.error("Try this again when the debugger is stopped inside an agent")
            return

        arg = f"{agent.__class__.__name__}:{ arg } {agent.unique_id}"
        return self.do_break_state(arg)

    do_bss = do_break_state_self


debugger = None


def set_trace(frame=None, **kwargs):
    global debugger
    if debugger is None:
        debugger = Debug(**kwargs)
    frame = frame or sys._getframe().f_back
    debugger.set_trace(frame)


def post_mortem(traceback=None, **kwargs):
    global debugger
    if debugger is None:
        debugger = Debug(**kwargs)
    t = sys.exc_info()[2]
    debugger.reset()
    debugger.interaction(None, t)



from __future__ import annotations

from typing_extensions import Annotated
import annotated_types
from typing import *

from dataclasses import dataclass

class Parameter:
    pass


def floatrange(
    *,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    multiple_of: Optional[float] = None,
) -> type[float]:
    return Annotated[ 
        float,
        annotated_types.Interval(gt=gt, ge=ge, lt=lt, le=le),
        annotated_types.MultipleOf(multiple_of) if multiple_of is not None else None,
    ]

function = Annotated[Callable, Parameter]
Integer = Annotated[int, Parameter]
Float = Annotated[float, Parameter]


probability = floatrange(ge=0, le=1)


import argparse
from server import ModularServer
from simulator import Simulator


def run(simulator, name="SOIL", port=8001, verbose=False):
    server = ModularServer(
        simulator, name=(name[0] if isinstance(name, list) else name), verbose=verbose
    )
    server.port = port
    server.launch()


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description="Visualization of a Graph Model")

    parser.add_argument(
        "--name", "-n", nargs=1, default="SOIL", help="name of the simulation"
    )
    parser.add_argument(
        "--dump", "-d", help="dumping results in folder output", action="store_true"
    )
    parser.add_argument(
        "--port", "-p", nargs=1, default=8001, help="port for launching the server"
    )
    parser.add_argument("--verbose", "-v", help="verbose mode", action="store_true")
    args = parser.parse_args()

    soil = Simulator(dump=args.dump)
    run(
        soil,
        name=args.name,
        port=(args.port[0] if isinstance(args.port, list) else args.port),
        verbose=args.verbose,
    )



import numpy as np
from hashlib import sha512
from . import Agent, state, default_state


class SISaModel(Agent):
    """
    Settings:
        neutral_discontent_spon_prob

        neutral_discontent_infected_prob

        neutral_content_spon_prob

        neutral_content_infected_prob

        discontent_neutral

        discontent_content

        variance_d_c

        content_discontent

        variance_c_d

        content_neutral

        standard_variance
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        seed = self.model._seed
        if isinstance(seed, (str, bytes, bytearray)):
            if isinstance(seed, str):
                seed = seed.encode()
            seed = int.from_bytes(seed + sha512(seed).digest(), 'big')

        random = np.random.default_rng(seed=seed)

        self.neutral_discontent_spon_prob = random.normal(
            self.model.neutral_discontent_spon_prob, self.model.standard_variance
        )
        self.neutral_discontent_infected_prob = random.normal(
            self.model.neutral_discontent_infected_prob, self.model.standard_variance
        )
        self.neutral_content_spon_prob = random.normal(
            self.model.neutral_content_spon_prob, self.model.standard_variance
        )
        self.neutral_content_infected_prob = random.normal(
            self.model.neutral_content_infected_prob, self.model.standard_variance
        )

        self.discontent_neutral = random.normal(
            self.model.discontent_neutral, self.model.standard_variance
        )
        self.discontent_content = random.normal(
            self.model.discontent_content, self.model.variance_d_c
        )

        self.content_discontent = random.normal(
            self.model.content_discontent, self.model.variance_c_d
        )
        self.content_neutral = random.normal(
            self.model.discontent_neutral, self.model.standard_variance
        )

    @default_state
    @state
    def neutral(self):
        # Spontaneous effects
        if self.prob(self.neutral_discontent_spon_prob):
            return self.discontent
        if self.prob(self.neutral_content_spon_prob):
            return self.content

        # Infected
        discontent_neighbors = self.count_neighbors(state_id=self.discontent)
        if self.prob(discontent_neighbors * self.neutral_discontent_infected_prob):
            return self.discontent
        content_neighbors = self.count_neighbors(state_id=self.content.id)
        if self.prob(content_neighbors * self.neutral_content_infected_prob):
            return self.content
        return self.neutral

    @state
    def discontent(self):
        # Healing
        if self.prob(self.discontent_neutral):
            return self.neutral

        # Superinfected
        content_neighbors = self.count_neighbors(state_id=self.content.id)
        if self.prob(content_neighbors * self.discontent_content):
            return self.content
        return self.discontent

    @state
    def content(self):
        # Healing
        if self.prob(self.content_neutral):
            return self.neutral

        # Superinfected
        discontent_neighbors = self.count_neighbors(state_id=self.discontent.id)
        if self.prob(discontent_neighbors * self.content_discontent):
            self.discontent
        return self.content
